# #!/usr/bin/env python
# # =============================================================================#
# #                                                                             #
# # NAME:     do_RM-clean.py                                                    #
# #                                                                             #
# # PURPOSE:  Run RM-clean on a  cube of dirty Faraday dispersion functions.    #
# #                                                                             #
# # MODIFIED: 15-May-2016 by C. Purcell                                         #
# # MODIFIED: 23-October-2019 by A. Thomson                                     #
# #                                                                             #
# # =============================================================================#
# #                                                                             #
# # The MIT License (MIT)                                                       #
# #                                                                             #
# # Copyright (c) 2015 Cormac R. Purcell                                        #
# #                                                                             #
# # Permission is hereby granted, free of charge, to any person obtaining a     #
# # copy of this software and associated documentation files (the "Software"),  #
# # to deal in the Software without restriction, including without limitation   #
# # the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# # and/or sell copies of the Software, and to permit persons to whom the       #
# # Software is furnished to do so, subject to the following conditions:        #
# #                                                                             #
# # The above copyright notice and this permission notice shall be included in  #
# # all copies or substantial portions of the Software.                         #
# #                                                                             #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# # DEALINGS IN THE SOFTWARE.                                                   #
# #                                                                             #
# # =============================================================================#

# import os
# import sys
# import time

# import astropy.io.fits as pf
# import numpy as np

# try:
#     import schwimmbad

#     parallel_available = True
# except:
#     parallel_available = False

# from rm_lite.tools_3d.rmsynth import _setStokes
# from rm_lite.utils.synthesis import do_rmclean_hogbom, fits_make_lin_axis

# C = 2.997924538e8  # Speed of light [m/s]


# # -----------------------------------------------------------------------------#
# def run_rmclean(
#     fitsFDF,
#     fitsRMSF,
#     cutoff,
#     maxIter=1000,
#     gain=0.1,
#     nBits=32,
#     pool=None,
#     chunksize=None,
#     verbose=True,
#     log=print,
#     window=np.nan,
# ):
#     """Run RM-CLEAN on a 2/3D FDF cube given an RMSF cube stored as FITS.

#     If you want to run RM-CLEAN on arrays, just use util_RM.do_rmclean_hogbom.

#     Args:
#         fitsFDF (str): Name of FDF FITS file.
#         fitsRMSF (str): Name of RMSF FITS file
#         cutoff (float): CLEAN cutoff in flux units

#     Kwargs:
#         maxIter (int): Maximum number of CLEAN iterations per pixel.
#         gain (float): CLEAN loop gain.
#         nBits (int): Precision of floating point numbers.
#         pool (multiprocessing Pool): Pool function from multiprocessing
#             or schwimmbad
#         chunksize (int): Number of chunks which it submits to the process pool
#             as separate tasks. The (approximate) size of these chunks can be
#             specified by setting chunksize to a positive integer.
#         verbose (bool): Verbosity.
#         log (function): Which logging function to use.

#     Returns:
#         cleanFDF (ndarray): Cube of RMCLEANed FDFs.
#         ccArr (ndarray): Cube of RMCLEAN components (i.e. the model).
#         iterCountArr (ndarray): Cube of number of RMCLEAN iterations.
#         residFDF (ndarray): Cube of residual RMCLEANed FDFs.
#         head (fits.header): Header of FDF FITS file for template.

#     """

#     # Default data types
#     dtFloat = "float" + str(nBits)
#     dtComplex = "complex" + str(2 * nBits)

#     # Read the FDF
#     dirtyFDF, head, FD_axis = read_FDF_cubes(fitsFDF)

#     phiArr_radm2 = fits_make_lin_axis(head, axis=FD_axis - 1, dtype=dtFloat)

#     # Read the RMSF

#     RMSFArr, headRMSF, FD_axis = read_FDF_cubes(fitsRMSF)
#     HDULst = pf.open(
#         fitsRMSF.replace("_real", "_FWHM")
#         .replace("_im", "_FWHM")
#         .replace("_tot", "_FWHM"),
#         "readonly",
#         memmap=True,
#     )
#     fwhmRMSFArr = np.squeeze(HDULst[0].data)
#     HDULst.close()
#     phi2Arr_radm2 = fits_make_lin_axis(headRMSF, axis=FD_axis - 1, dtype=dtFloat)

#     startTime = time.time()

#     # Do the clean
#     cleanFDF, ccArr, iterCountArr, residFDF = do_rmclean_hogbom(
#         dirtyFDF=dirtyFDF,
#         phiArr_radm2=phiArr_radm2,
#         RMSFArr=RMSFArr,
#         phi2Arr_radm2=phi2Arr_radm2,
#         fwhmRMSFArr=fwhmRMSFArr,
#         cutoff=cutoff,
#         maxIter=maxIter,
#         gain=gain,
#         verbose=verbose,
#         doPlots=False,
#         pool=pool,
#         chunksize=chunksize,
#         window=window,
#     )

#     endTime = time.time()
#     cputime = endTime - startTime
#     if verbose:
#         log("> RM-clean completed in %.2f seconds." % cputime)
#     if verbose:
#         log("Saving the clean FDF and ancillary FITS files")

#     # Move FD axis back to original position, and restore dimensionality:
#     old_Ndim, FD_axis = find_axes(head)
#     new_Ndim = cleanFDF.ndim
#     # The difference is the number of dimensions that need to be added:
#     if old_Ndim - new_Ndim != 0:  # add missing (degenerate) dimensions back in
#         cleanFDF = np.expand_dims(cleanFDF, axis=tuple(range(old_Ndim - new_Ndim)))
#         ccArr = np.expand_dims(ccArr, axis=tuple(range(old_Ndim - new_Ndim)))
#         residFDF = np.expand_dims(residFDF, axis=tuple(range(old_Ndim - new_Ndim)))
#     # New dimensions are added to the beginning of the axis ordering
#     # (reverse of FITS ordering)

#     # Move the FDF axis to it's original spot. Hopefully this means that all
#     # axes are in their original place after all of that.
#     cleanFDF = np.moveaxis(cleanFDF, old_Ndim - new_Ndim, old_Ndim - FD_axis)
#     ccArr = np.moveaxis(ccArr, old_Ndim - new_Ndim, old_Ndim - FD_axis)
#     residFDF = np.moveaxis(residFDF, old_Ndim - new_Ndim, old_Ndim - FD_axis)

#     return cleanFDF, ccArr, iterCountArr, residFDF, head


# def writefits(
#     cleanFDF,
#     ccArr,
#     iterCountArr,
#     residFDF,
#     headtemp,
#     nBits=32,
#     prefixOut="",
#     outDir="",
#     write_separate_FDF=False,
#     verbose=True,
#     log=print,
# ):
#     """Write data to disk in FITS


#     Output files:
#         Default:
#             FDF_clean.fits: RMCLEANed FDF, in 3 extensions: Q,U, and PI.
#             FDF_CC.fits: RMCLEAN components, in 3 extensions: Q,U, and PI.
#             CLEAN_nIter.fits: RMCLEAN iterations.

#         write_separate_FDF=True:
#             FDF_clean_real.fits and FDF_CC.fits are split into
#             three constituent components:
#                 FDF_clean_real.fits: Stokes Q
#                 FDF_clean_im.fits: Stokes U
#                 FDF_clean_tot.fits: Polarized Intensity (sqrt(Q^2+U^2))
#                 FDF_CC_real.fits: Stokes Q
#                 FDF_CC_im.fits: Stokes U
#                 FDF_CC_tot.fits: Polarized Intensity (sqrt(Q^2+U^2))
#                 CLEAN_nIter.fits: RMCLEAN iterations.
#     Args:
#         cleanFDF (ndarray): Cube of RMCLEANed FDFs.
#         ccArr (ndarray): Cube of RMCLEAN components (i.e. the model).
#         iterCountArr (ndarray): Cube of number of RMCLEAN iterations.
#         residFDF (ndarray): Cube of residual RMCLEANed FDFs.

#     Kwargs:
#         prefixOut (str): Prefix for filenames.
#         outDir (str): Directory to save files.
#         write_separate_FDF (bool): Write Q, U, and PI separately?
#         verbose (bool): Verbosity.
#         log (function): Which logging function to use.
#     """
#     # Default data types
#     dtFloat = "float" + str(nBits)
#     dtComplex = "complex" + str(2 * nBits)
#     header = headtemp.copy()

#     header["HISTORY"] = "RM-CLEAN3D output from RM-Tools"

#     if outDir == "":  # To prevent code breaking if file is in current directory
#         outDir = "."
#     # Save the clean FDF
#     if not write_separate_FDF:
#         header = _setStokes(header, "Q")
#         hdu0 = pf.PrimaryHDU(cleanFDF.real.astype(dtFloat), header)
#         header = _setStokes(header, "U")
#         hdu1 = pf.ImageHDU(cleanFDF.imag.astype(dtFloat), header)
#         header = _setStokes(
#             header, "PI"
#         )  # Sets Stokes axis to zero, which is a non-standard value.
#         del header["STOKES"]
#         hdu2 = pf.ImageHDU(np.abs(cleanFDF).astype(dtFloat), header)

#         fitsFileOut = outDir + "/" + prefixOut + "FDF_clean.fits"
#         if verbose:
#             log("> %s" % fitsFileOut)
#         hduLst = pf.HDUList([hdu0, hdu1, hdu2])
#         hduLst.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#         hduLst.close()
#     else:
#         header = _setStokes(header, "Q")
#         hdu0 = pf.PrimaryHDU(cleanFDF.real.astype(dtFloat), header)
#         header = _setStokes(header, "U")
#         hdu1 = pf.PrimaryHDU(cleanFDF.imag.astype(dtFloat), header)
#         header = _setStokes(
#             header, "PI"
#         )  # Sets Stokes axis to zero, which is a non-standard value.
#         del header["STOKES"]
#         hdu2 = pf.PrimaryHDU(np.abs(cleanFDF).astype(dtFloat), header)

#         fitsFileOut = outDir + "/" + prefixOut + "FDF_clean_real.fits"
#         hdu0.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#         if verbose:
#             log("> %s" % fitsFileOut)
#         fitsFileOut = outDir + "/" + prefixOut + "FDF_clean_im.fits"
#         hdu1.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#         if verbose:
#             log("> %s" % fitsFileOut)
#         fitsFileOut = outDir + "/" + prefixOut + "FDF_clean_tot.fits"
#         hdu2.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#         if verbose:
#             log("> %s" % fitsFileOut)

#     if not write_separate_FDF:
#         # Save the complex clean components as another file.
#         header = _setStokes(header, "Q")
#         hdu0 = pf.PrimaryHDU(ccArr.real.astype(dtFloat), header)
#         header = _setStokes(header, "U")
#         hdu1 = pf.ImageHDU(ccArr.imag.astype(dtFloat), header)
#         header = _setStokes(
#             header, "PI"
#         )  # Sets Stokes axis to zero, which is a non-standard value.
#         del header["STOKES"]
#         hdu2 = pf.ImageHDU(np.abs(ccArr).astype(dtFloat), header)

#         fitsFileOut = outDir + "/" + prefixOut + "FDF_CC.fits"
#         if verbose:
#             log("> %s" % fitsFileOut)
#         hduLst = pf.HDUList([hdu0, hdu1, hdu2])
#         hduLst.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#         hduLst.close()
#     else:
#         header = _setStokes(header, "Q")
#         hdu0 = pf.PrimaryHDU(ccArr.real.astype(dtFloat), header)
#         header = _setStokes(header, "U")
#         hdu1 = pf.PrimaryHDU(ccArr.imag.astype(dtFloat), header)
#         header = _setStokes(
#             header, "PI"
#         )  # Sets Stokes axis to zero, which is a non-standard value.
#         del header["STOKES"]
#         hdu2 = pf.PrimaryHDU(np.abs(ccArr).astype(dtFloat), header)

#         fitsFileOut = outDir + "/" + prefixOut + "FDF_CC_real.fits"
#         hdu0.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#         if verbose:
#             log("> %s" % fitsFileOut)
#         fitsFileOut = outDir + "/" + prefixOut + "FDF_CC_im.fits"
#         hdu1.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#         if verbose:
#             log("> %s" % fitsFileOut)
#         fitsFileOut = outDir + "/" + prefixOut + "FDF_CC_tot.fits"
#         hdu2.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#         if verbose:
#             log("> %s" % fitsFileOut)

#     # Because there can be problems with different axes having different FITS keywords,
#     # don't try to remove the FD axis, but just make it degenerate.

#     if headtemp["NAXIS"] > 2:
#         header["NAXIS3"] = 1
#         header["CTYPE3"] = ("DEGENERATE", "Axis left in to avoid FITS errors")
#         header["CUNIT3"] = ""
#     if headtemp["NAXIS"] == 4:
#         header["NAXIS4"] = 1
#         header["CTYPE4"] = ("DEGENERATE", "Axis left in to avoid FITS errors")
#         header["CUNIT4"] = ""

#     # Save the iteration count mask
#     fitsFileOut = outDir + "/" + prefixOut + "CLEAN_nIter.fits"
#     if verbose:
#         log("> %s" % fitsFileOut)
#     header["BUNIT"] = "Iterations"
#     if "STOKES" in header:
#         del header["STOKES"]
#     hdu0 = pf.PrimaryHDU(
#         np.expand_dims(
#             iterCountArr.astype(dtFloat),
#             axis=tuple(range(headtemp["NAXIS"] - iterCountArr.ndim)),
#         ),
#         header,
#     )
#     hduLst = pf.HDUList([hdu0])
#     hduLst.writeto(fitsFileOut, output_verify="fix", overwrite=True)
#     hduLst.close()


# # Old method (for multi-extension files)
# def find_axes(header):
#     """Identifies how many axes are present in a FITS file, and which is the
#     Faraday depth axis. Necessary for bookkeeping on cube dimensionality,
#     given that RM-clean only supports 3D cubes, but data may be 4D files."""
#     Ndim = header["NAXIS"]
#     FD_axis = Ndim
#     # Check for FD axes:
#     for i in range(1, Ndim + 1):
#         try:
#             if "FDEP" in header["CTYPE" + str(i)].upper():
#                 FD_axis = i
#         except:
#             pass  # The try statement is needed for if the FITS header does not
#             # have CTYPE keywords.
#     return Ndim, FD_axis


# def read_FDF_cubes(filename):
#     """Read in a FDF/RMSF cube. Input filename can be any of real, imag, or tot components.
#     Figures out which axis is Faraday depth and
#     puts it first (in numpy order) to accommodate the rest of the code.
#     Returns: (complex_cube, header,FD_axis)
#     """
#     HDUreal = pf.open(
#         filename.replace("_tot", "_real").replace("_im", "_real"),
#         "readonly",
#         memmap=True,
#     )
#     head = HDUreal[0].header.copy()
#     FDFreal = HDUreal[0].data

#     HDUimag = pf.open(
#         filename.replace("_tot", "_im").replace("_real", "_im"), "readonly", memmap=True
#     )
#     FDFimag = HDUimag[0].data
#     complex_cube = FDFreal + 1j * FDFimag

#     # Identify Faraday depth axis (assumed to be last one if not explicitly found)
#     Ndim, FD_axis = find_axes(head)

#     # Move FD axis to first place in numpy order.
#     if FD_axis != Ndim:
#         complex_cube = np.moveaxis(complex_cube, Ndim - FD_axis, 0)

#     # Remove degenerate axes to prevent problems with later steps.
#     complex_cube = complex_cube.squeeze()

#     return complex_cube, head, FD_axis


# # -----------------------------------------------------------------------------#
# def main():
#     import argparse

#     """
#     Start the function to perform RM-clean if called from the command line.
#     """

#     # Help string to be shown using the -h option
#     descStr = """
#     Run RM-CLEAN on a cube of Faraday dispersion functions (FDFs), applying
#     a cube of rotation measure spread functions created by the script
#     'do_RMsynth_3D.py'. Saves cubes of deconvolved FDFs & clean-component
#     spectra, and a pixel map showing the number of iterations performed.
#     Set any of the multiprocessing options to enable parallelization
#     (otherwise, pixels will be processed serially).

#     Expects that the input is in the form of the Stokes-separated
#     (single extension) FITS cubes produced by do_RMsynth_3D.
#     """

#     # Parse the command line options
#     parser = argparse.ArgumentParser(
#         description=descStr, formatter_class=argparse.RawTextHelpFormatter
#     )
#     parser.add_argument(
#         "fitsFDF",
#         metavar="FDF_dirty.fits",
#         nargs=1,
#         help="FITS cube containing the dirty FDF.\n(Can be any of the FDF output cubes from do_RMsynth_3D.py)",
#     )
#     parser.add_argument(
#         "fitsRMSF",
#         metavar="RMSF.fits",
#         nargs=1,
#         help="FITS cube containing the RMSF and FWHM image.\n(Cans be any of the RMSF output cubes from do_RMsynth_3D.py)",
#     )
#     parser.add_argument(
#         "-c",
#         dest="cutoff",
#         type=float,
#         default=1,
#         help="Initial CLEAN cutoff in flux units [1].",
#     )
#     parser.add_argument(
#         "-w",
#         dest="window",
#         type=float,
#         default=np.nan,
#         help="Threshold for (deeper) windowed clean [Not used if not set].",
#     )
#     parser.add_argument(
#         "-n",
#         dest="maxIter",
#         type=int,
#         default=1000,
#         help="Maximum number of CLEAN iterations per pixel [1000].",
#     )
#     parser.add_argument(
#         "-g", dest="gain", type=float, default=0.1, help="CLEAN loop gain [0.1]."
#     )
#     parser.add_argument(
#         "-o",
#         dest="prefixOut",
#         default="",
#         help="Prefix to prepend to output files [None].",
#     )
#     parser.add_argument(
#         "-f",
#         dest="write_separate_FDF",
#         action="store_false",
#         help="Store different Stokes as FITS extensions [False, store as separate files].",
#     )

#     parser.add_argument(
#         "-v", dest="verbose", action="store_true", help="Verbose [False]."
#     )

#     group = parser.add_mutually_exclusive_group()
#     group.add_argument(
#         "--ncores",
#         dest="n_cores",
#         default=1,
#         type=int,
#         help="Number of processes (uses multiprocessing).",
#     )
#     parser.add_argument(
#         "--chunk",
#         dest="chunk",
#         default=None,
#         type=int,
#         help="Chunk size (uses multiprocessing -- not available in MPI!)",
#     )
#     group.add_argument(
#         "--mpi", dest="mpi", default=False, action="store_true", help="Run with MPI."
#     )

#     args = parser.parse_args()

#     # Check if the user is trying to use parallelization without installing it:
#     if parallel_available == False and (
#         args.n_cores != 1 or args.chunk != None or args.mpi != False
#     ):
#         raise Exception(
#             "Parallel processing not available. Please install the schwimmbad module to enable parallel processing."
#         )

#     # If parallelization requested use it, otherwise use the old-fashioned way.
#     if parallel_available == True and (
#         args.n_cores != 1 or args.chunk != None or args.mpi != False
#     ):
#         pool = schwimmbad.choose_pool(mpi=args.mpi, processes=args.n_cores)
#         if args.mpi:
#             if not pool.is_master():
#                 pool.wait()
#                 sys.exit(0)
#         if args.n_cores > 1:
#             chunksize = args.chunk
#         else:
#             chunksize = None
#     else:
#         pool = None
#         chunksize = None

#     verbose = args.verbose
#     # Sanity checks
#     for f in args.fitsFDF + args.fitsRMSF:
#         if not os.path.exists(f):
#             print("File does not exist: '%s'." % f)
#             sys.exit()
#     (dataDir,) = os.path.split(args.fitsFDF[0])

#     # Run RM-CLEAN on the cubes
#     cleanFDF, ccArr, iterCountArr, residFDF, headtemp = run_rmclean(
#         fitsFDF=args.fitsFDF[0],
#         fitsRMSF=args.fitsRMSF[0],
#         cutoff=args.cutoff,
#         maxIter=args.maxIter,
#         gain=args.gain,
#         chunksize=chunksize,
#         nBits=32,
#         window=args.window,
#         verbose=verbose,
#     )
#     # Write results to disk
#     writefits(
#         cleanFDF,
#         ccArr,
#         iterCountArr,
#         residFDF,
#         headtemp,
#         prefixOut=args.prefixOut,
#         outDir=dataDir,
#         write_separate_FDF=args.write_separate_FDF,
#         verbose=verbose,
#     )


# # -----------------------------------------------------------------------------#
# if __name__ == "__main__":
#     main()



from __future__ import annotations

import numpy as np
from rm_lite.utils.arrays import arange, nd_to_two_d, two_d_to_nd


def test_nd_to_two_d():
    array_1d = np.arange(288)
    array_2d = array_1d.reshape(36, 8)
    array_3d = array_1d.reshape(36, 2, 4)
    array_4d = array_1d.reshape(36, 2, 2, 2)

    # test shapes
    # First axis is kept intact
    assert nd_to_two_d(array_1d).shape == (288, 1)
    assert nd_to_two_d(array_2d).shape == (36, 8)
    assert nd_to_two_d(array_3d).shape == (36, 8)
    assert nd_to_two_d(array_4d).shape == (36, 8)


def test_two_d_to_nd():
    array_1d = np.arange(12)
    array_2d = array_1d.reshape(3, 4)
    array_3d = array_1d.reshape(2, 3, 2)

    assert two_d_to_nd(nd_to_two_d(array_1d), array_1d.shape).shape == array_1d.shape
    assert two_d_to_nd(nd_to_two_d(array_2d), array_2d.shape).shape == array_2d.shape
    assert two_d_to_nd(nd_to_two_d(array_3d), array_3d.shape).shape == array_3d.shape

    # test round trip
    assert np.array_equal(two_d_to_nd(nd_to_two_d(array_1d), array_1d.shape), array_1d)
    assert np.array_equal(two_d_to_nd(nd_to_two_d(array_2d), array_2d.shape), array_2d)
    assert np.array_equal(two_d_to_nd(nd_to_two_d(array_3d), array_3d.shape), array_3d)


def test_arange():
    paras_minimal_working_example = {
        "arange simple": {
            "start": 0,
            "stop": 7,
            "step": 1,
            "include_start": True,
            "include_stop": False,
            "res_exp": np.array([0, 1, 2, 3, 4, 5, 6]),
        },
        "stop not on grid": {
            "start": 0,
            "stop": 6.5,
            "step": 1,
            "include_start": True,
            "include_stop": False,
            "res_exp": np.array([0, 1, 2, 3, 4, 5, 6]),
        },
        "arange failing example: stop excl": {
            "start": 1,
            "stop": 1.3,
            "step": 0.1,
            "include_start": True,
            "include_stop": False,
            "res_exp": np.array([1.0, 1.1, 1.2]),
        },
        "arange failing example: stop incl": {
            "start": 1,
            "stop": 1.3,
            "step": 0.1,
            "include_start": True,
            "include_stop": True,
            "res_exp": np.array([1.0, 1.1, 1.2, 1.3]),
        },
        "arange failing example: stop excl + start excl": {
            "start": 1,
            "stop": 1.3,
            "step": 0.1,
            "include_start": False,
            "include_stop": False,
            "res_exp": np.array([1.1, 1.2]),
        },
        "arange failing example: stop incl + start excl": {
            "start": 1,
            "stop": 1.3,
            "step": 0.1,
            "include_start": False,
            "include_stop": True,
            "res_exp": np.array([1.1, 1.2, 1.3]),
        },
    }
    for desc, paras in paras_minimal_working_example.items():
        start, stop, step, include_start, include_stop, res_exp = paras.values()
        res = arange(
            start, stop, step, include_start=include_start, include_stop=include_stop
        )
        assert np.allclose(res, res_exp), (
            f"Unexpected result in {desc}: {res=}, {res_exp=}"
        )



"""Array utils"""

from __future__ import annotations

import logging
from typing import TypeVar

import numpy as np
from numpy.typing import NDArray

from rm_lite.utils.logging import TqdmToLogger, logger

TQDM_OUT = TqdmToLogger(logger, level=logging.INFO)

DType = TypeVar("DType", bound=np.generic)


def nd_to_two_d(arr: NDArray[DType]) -> NDArray[DType]:
    """Convert an array to 2D.

    - If arr is 1D, it will be reshaped as a column vector (shape: (N, 1)).
    - If arr is already 2D, it is returned as is.
    - If arr has more than 2 dimensions, the first axis is kept intact
      and all remaining axes are flattened. For example, an array with
      shape (a, b, c, d) will become shape (a, b*c*d).

    Args:
        arr (NDArray[Any]): N-dimensional array.

    Returns:
        NDArray[Any]: 2D array.
    """
    arr = np.asarray(arr)
    if arr.ndim == 0:
        return arr.reshape(1, 1)
    if arr.ndim == 1:
        return arr.reshape(-1, 1)
    if arr.ndim == 2:
        return arr
    return arr.reshape(arr.shape[0], -1)


def two_d_to_nd(
    arr2d: NDArray[DType], original_shape: tuple[int, ...]
) -> NDArray[DType]:
    """
    Reverse the to_2d operation.

    Parameters:
        arr2d (array-like): the 2D array (result from to_2d).
        original_shape (tuple): the shape of the original array before flattening.

    Returns:
        The array reshaped back to its original shape.

    The function assumes:
        - For an original 1D array, original_shape is (N,). In this case, arr2d is of shape (N, 1)
          and will be flattened back to (N,).
        - For an original 2D array, original_shape is (M, N) and arr2d is already (M, N).
        - For an original N_D array (N_D > 2) with shape (a, b, c, ...), arr2d is assumed to have shape
          (a, b*c*...) and will be reshaped back to (a, b, c, ...).
    """
    arr2d = np.asarray(arr2d)
    # If the original was 1D, simply flatten the second axis.
    if len(original_shape) == 1:
        # (N,1) -> (N,)
        return arr2d.ravel()
    # If the original was 2D, reshape directly.
    if len(original_shape) == 2:
        return arr2d.reshape(original_shape)

    # For N_D arrays (with ndim > 2), the to_2d function preserved the first axis
    # and flattened the remaining dimensions.
    expected_first_dim = original_shape[0]
    expected_rest = int(np.prod(original_shape[1:]))
    if arr2d.shape != (expected_first_dim, expected_rest):
        msg = "The provided original shape is not consistent with the 2D array shape."
        raise ValueError(msg)
    return arr2d.reshape(original_shape)


# from https://stackoverflow.com/questions/50299172/range-or-numpy-arange-with-end-limit-include
def arange(
    start: float | int,
    stop: float | int,
    step: float | int,
    rtol: float = 1e-05,
    atol: float = 1e-08,
    include_start: bool = True,
    include_stop: bool = False,
    **kwargs,
) -> NDArray[np.float64]:
    """
    Combines numpy.arange and numpy.isclose to mimic open, half-open and closed intervals.

    Avoids also floating point rounding errors as with
    >>> np.arange(1, 1.3, 0.1)
    array([1., 1.1, 1.2, 1.3])


    Args:
        start (float | int): Start of the interval.
        stop (float | int): End of the interval.
        step (float | int): Spacing between values.
        rtol (float, optional): if last element of array is within this relative tolerance to stop and include[0]==False, it is skipped. Defaults to 1e-05.
        atol (float, optional): if last element of array is within this relative tolerance to stop and include[1]==False, it is skipped. Defaults to 1e-08.
        include_start (bool, optional): if first element is included in the returned array. Defaults to True.
        include_stop (bool, optional): if last elements are included in the returned array if stop equals last element. Defaults to False.
        kwargs: passed to np.arange

    Returns:
        _type_: as np.arange but eventually with first and last element stripped/added
    """
    arr = np.arange(start, stop, step, **kwargs)
    if not include_start:
        arr = np.delete(arr, 0)

    if include_stop:
        if np.isclose(arr[-1] + step, stop, rtol=rtol, atol=atol):
            # arr = np.c_[arr, arr[-1] + step]
            arr = np.append(arr, arr[-1] + step)
    elif np.isclose(arr[-1], stop, rtol=rtol, atol=atol):
        arr = np.delete(arr, -1)
    return arr



"""Tests for the NUFFT sections"""

from __future__ import annotations

import contextlib
from time import time
from typing import NamedTuple

import numpy as np
from numpy.typing import NDArray
from rm_lite.utils.logging import logger
from rm_lite.utils.synthesis import (
    get_rmsf_nufft,
    make_double_phi_arr,
    rmsynth_nufft,
)
from tqdm import trange


class FakeData(NamedTuple):
    """Fake data for testing RM-synthesis"""

    freq: NDArray[np.float64]
    """Frequency array (Hz)"""
    lsq: NDArray[np.float64]
    """Wavelength^2 array (m^2)"""
    rm: float
    """Rotation Measure (rad/m^2)"""
    stokes_q: NDArray[np.float64]
    """Stokes Q array"""
    stokes_u: NDArray[np.float64]
    """Stokes U array"""
    weights: NDArray[np.float64]
    """Weights array"""
    lsq_0: float
    """Weighted mean of lsq"""
    phis: NDArray[np.float64]
    """Faraday depth array (rad/m^2)"""


def do_rmsynth_planes_old(
    dataQ,
    dataU,
    lambdaSqArr_m2,
    phiArr_radm2,
    weightArr=None,
    lam0Sq_m2=None,
    nBits=32,
    verbose=False,
    log=print,
):
    """Perform RM-synthesis on Stokes Q and U cubes (1,2 or 3D). This version
    of the routine loops through spectral planes and is faster than the pixel-
    by-pixel code. This version also correctly deals with isolated clumps of
    NaN-flagged voxels within the data-cube (unlikely in interferometric cubes,
    but possible in single-dish cubes). Input data must be in standard python
    [z,y,x] order, where z is the frequency axis in ascending order.

    dataQ           ... 1, 2 or 3D Stokes Q data array
    dataU           ... 1, 2 or 3D Stokes U data array
    lambdaSqArr_m2  ... vector of wavelength^2 values (ascending freq order)
    phiArr_radm2    ... vector of trial Faraday depth values
    weightArr       ... vector of weights, default [None] is Uniform (all 1s)
    nBits           ... precision of data arrays [32]
    verbose         ... print feedback during calculation [False]
    log             ... function to be used to output messages [print]

    """

    # Default data types
    dtFloat = "float" + str(nBits)
    dtComplex = "complex" + str(2 * nBits)

    # Set the weight array
    if weightArr is None:
        weightArr = np.ones(lambdaSqArr_m2.shape, dtype=dtFloat)
    weightArr = np.where(np.isnan(weightArr), 0.0, weightArr)

    # Sanity check on array sizes
    if weightArr.shape != lambdaSqArr_m2.shape:
        log("Err: Lambda^2 and weight arrays must be the same shape.")
        return None, None
    if dataQ.shape != dataU.shape:
        log("Err: Stokes Q and U data arrays must be the same shape.")
        return None, None
    nDims = len(dataQ.shape)
    if not nDims <= 3:
        log("Err: data dimensions must be <= 3.")
        return None, None
    if dataQ.shape[0] != lambdaSqArr_m2.shape[0]:
        log(
            f"Err: Data depth does not match lambda^2 vector ({dataQ.shape[0]} vs {lambdaSqArr_m2.shape[0]}).",
            end=" ",
        )
        log("     Check that data is in [z, y, x] order.")
        return None, None

    # Reshape the data arrays to 3 dimensions
    if nDims == 1:
        dataQ = np.reshape(dataQ, (dataQ.shape[0], 1, 1))
        dataU = np.reshape(dataU, (dataU.shape[0], 1, 1))
    elif nDims == 2:
        dataQ = np.reshape(dataQ, (dataQ.shape[0], dataQ.shape[1], 1))
        dataU = np.reshape(dataU, (dataU.shape[0], dataU.shape[1], 1))

    # Create a complex polarised cube, B&dB Eqns. (8) and (14)
    # Array has dimensions [nFreq, nY, nX]
    pCube = (dataQ + 1j * dataU) * weightArr[:, np.newaxis, np.newaxis]

    # Check for NaNs (flagged data) in the cube & set to zero
    mskCube = np.isnan(pCube)
    pCube = np.nan_to_num(pCube)

    # If full planes are flagged then set corresponding weights to zero
    mskPlanes = np.sum(np.sum(~mskCube, axis=1), axis=1)
    mskPlanes = np.where(mskPlanes == 0, 0, 1)
    weightArr *= mskPlanes

    # Initialise the complex Faraday Dispersion Function cube
    nX = dataQ.shape[-1]
    nY = dataQ.shape[-2]
    nPhi = phiArr_radm2.shape[0]
    FDFcube = np.zeros((nPhi, nY, nX), dtype=dtComplex)

    # lam0Sq_m2 is the weighted mean of lambda^2 distribution (B&dB Eqn. 32)
    # Calculate a global lam0Sq_m2 value, ignoring isolated flagged voxels
    K = 1.0 / np.sum(weightArr)
    if lam0Sq_m2 is None:
        lam0Sq_m2 = K * np.sum(weightArr * lambdaSqArr_m2)
    if not np.isfinite(lam0Sq_m2):  # Can happen if all channels are NaNs/zeros
        lam0Sq_m2 = 0.0

    # The K value used to scale each FDF spectrum must take into account
    # flagged voxels data in the datacube and can be position dependent
    weightCube = np.invert(mskCube) * weightArr[:, np.newaxis, np.newaxis]
    with np.errstate(divide="ignore", invalid="ignore"):
        KArr = np.true_divide(1.0, np.sum(weightCube, axis=0))
        KArr[KArr == np.inf] = 0
        KArr = np.nan_to_num(KArr)

    # Do the RM-synthesis on each plane
    a = lambdaSqArr_m2 - lam0Sq_m2
    for i in trange(nPhi, desc="Running RM-synthesis by channel", disable=not verbose):
        arg = np.exp(-2.0j * phiArr_radm2[i] * a)[:, np.newaxis, np.newaxis]
        FDFcube[i, :, :] = KArr * np.sum(pCube * arg, axis=0)

    # Check for pixels that have Re(FDF)=Im(FDF)=0. across ALL Faraday depths
    # These pixels will be changed to NaN in the output
    zeromap = np.all(FDFcube == 0.0, axis=0)
    zeropxlist = np.where(zeromap)
    if np.shape(zeropxlist)[1] != 0:
        FDFcube[:, zeropxlist[0], zeropxlist[1]] = np.nan + 1.0j * np.nan

    # Remove redundant dimensions in the FDF array
    FDFcube = np.squeeze(FDFcube)

    return FDFcube, lam0Sq_m2


def extrap(x, xp, yp):
    """
    Wrapper to allow np.interp to linearly extrapolate at function ends.

    np.interp function with linear extrapolation
    http://stackoverflow.com/questions/2745329/how-to-make-scipy-interpolate
    -give-a-an-extrapolated-result-beyond-the-input-ran
    """

    y = np.interp(x, xp, yp)
    y = np.where(x < xp[0], yp[0] + (x - xp[0]) * (yp[0] - yp[1]) / (xp[0] - xp[1]), y)
    return np.where(
        x > xp[-1], yp[-1] + (x - xp[-1]) * (yp[-1] - yp[-2]) / (xp[-1] - xp[-2]), y
    )


def get_rmsf_planes_old(
    lambdaSqArr_m2,
    phiArr_radm2,
    weightArr=None,
    mskArr=None,
    lam0Sq_m2=None,
    double=True,
    nBits=32,
    verbose=False,
    log=print,
):
    """Calculate the Rotation Measure Spread Function from inputs. This version
    returns a cube (1, 2 or 3D) of RMSF spectra based on the shape of a
    boolean mask array, where flagged data are True and unflagged data False.
    If only whole planes (wavelength channels) are flagged then the RMSF is the
    same for all pixels and the calculation is done once and replicated to the
    dimensions of the mask. If some isolated voxels are flagged then the RMSF
    is calculated by looping through each wavelength plane, which can take some
    time. By default the routine returns the analytical width of the RMSF main
    lobe but can also use MPFIT to fit a Gaussian.

    lambdaSqArr_m2  ... vector of wavelength^2 values (ascending freq order)
    phiArr_radm2    ... vector of trial Faraday depth values
    weightArr       ... vector of weights, default [None] is no weighting
    maskArr         ... cube of mask values used to shape return cube [None]
    lam0Sq_m2       ... force a reference lambda^2 value (def=calculate) [None]
    double          ... pad the Faraday depth to double-size [True]
    fitRMSF         ... fit the main lobe of the RMSF with a Gaussian [False]
    fitRMSFreal     ... fit RMSF.real, rather than abs(RMSF) [False]
    nBits           ... precision of data arrays [32]
    verbose         ... print feedback during calculation [False]
    log             ... function to be used to output messages [print]

    """

    # Default data types
    dtFloat = "float" + str(nBits)
    dtComplex = "complex" + str(2 * nBits)

    # For cleaning the RMSF should extend by 1/2 on each side in phi-space
    phi2Arr = make_double_phi_arr(phiArr_radm2) if double else phiArr_radm2

    # Set the weight array
    if weightArr is None:
        weightArr = np.ones(lambdaSqArr_m2.shape, dtype=dtFloat)
    weightArr = np.where(np.isnan(weightArr), 0.0, weightArr)

    # Set the mask array (default to 1D, no masked channels)
    if mskArr is None:
        mskArr = np.zeros_like(lambdaSqArr_m2, dtype="bool")
        nDims = 1
    else:
        mskArr = mskArr.astype("bool")
        nDims = len(mskArr.shape)

    # Sanity checks on array sizes
    if weightArr.shape != lambdaSqArr_m2.shape:
        log("Err: wavelength^2 and weight arrays must be the same shape.")
        return None, None, None, None
    if not nDims <= 3:
        log("Err: mask dimensions must be <= 3.")
        return None, None, None, None
    if mskArr.shape[0] != lambdaSqArr_m2.shape[0]:
        log(
            f"Err: mask depth does not match lambda^2 vector ({mskArr.shape[0]} vs {lambdaSqArr_m2.shape[-1]}).",
            end=" ",
        )
        log("     Check that the mask is in [z, y, x] order.")
        return None, None, None, None

    # Reshape the mask array to 3 dimensions
    if nDims == 1:
        mskArr = np.reshape(mskArr, (mskArr.shape[0], 1, 1))
    elif nDims == 2:
        mskArr = np.reshape(mskArr, (mskArr.shape[0], mskArr.shape[1], 1))

    # Create a unit cube for use in RMSF calculation (negative of mask)
    # CVE: unit cube removed: it wasn't accurate for non-uniform weights, and was no longer used

    # Initialise the complex RM Spread Function cube
    nX = mskArr.shape[-1]
    nY = mskArr.shape[-2]
    nPix = nX * nY
    nPhi = phi2Arr.shape[0]
    RMSFcube = np.ones((nPhi, nY, nX), dtype=dtComplex)

    # If full planes are flagged then set corresponding weights to zero
    xySum = np.sum(np.sum(mskArr, axis=1), axis=1)
    mskPlanes = np.where(xySum == nPix, 0, 1)
    weightArr *= mskPlanes

    # Check for isolated clumps of flags (# flags in a plane not 0 or nPix)
    flagTotals = np.unique(xySum).tolist()
    with contextlib.suppress(Exception):
        flagTotals.remove(0)
    with contextlib.suppress(Exception):
        flagTotals.remove(nPix)
    do1Dcalc = True
    if len(flagTotals) > 0:
        do1Dcalc = False

    # lam0Sq is the weighted mean of LambdaSq distribution (B&dB Eqn. 32)
    # Calculate a single lam0Sq_m2 value, ignoring isolated flagged voxels
    K = 1.0 / np.nansum(weightArr)
    if lam0Sq_m2 is None:
        lam0Sq_m2 = K * np.nansum(weightArr * lambdaSqArr_m2)

    # Calculate the analytical FWHM width of the main lobe
    fwhmRMSF = 3.8 / (np.nanmax(lambdaSqArr_m2) - np.nanmin(lambdaSqArr_m2))

    # Do a simple 1D calculation and replicate along X & Y axes
    if do1Dcalc:
        if verbose:
            log("Calculating 1D RMSF and replicating along X & Y axes.")

        # Calculate the RMSF
        a = (-2.0 * 1j * phi2Arr).astype(dtComplex)
        b = lambdaSqArr_m2 - lam0Sq_m2
        RMSFArr = K * np.sum(weightArr * np.exp(np.outer(a, b)), 1)

        # Fit the RMSF main lobe
        fitStatus = -1

        # Replicate along X and Y axes
        RMSFcube = np.tile(RMSFArr[:, np.newaxis, np.newaxis], (1, nY, nX))
        fwhmRMSFArr = np.ones((nY, nX), dtype=dtFloat) * fwhmRMSF
        statArr = np.ones((nY, nX), dtype="int") * fitStatus

    # Calculate the RMSF at each pixel
    else:
        if verbose:
            log()

        # The K value used to scale each RMSF must take into account
        # isolated flagged voxels data in the datacube
        weightCube = np.invert(mskArr) * weightArr[:, np.newaxis, np.newaxis]
        with np.errstate(divide="ignore", invalid="ignore"):
            KArr = np.true_divide(1.0, np.sum(weightCube, axis=0))
            KArr[KArr == np.inf] = 0
            KArr = np.nan_to_num(KArr)

        # Calculate the RMSF for each plane
        a = lambdaSqArr_m2 - lam0Sq_m2
        for i in trange(nPhi, desc="Calculating RMSF by channel", disable=not verbose):
            arg = np.exp(-2.0j * phi2Arr[i] * a)[:, np.newaxis, np.newaxis]
            RMSFcube[i, :, :] = KArr * np.sum(weightCube * arg, axis=0)

        # Default to the analytical RMSF
        fwhmRMSFArr = np.ones((nY, nX), dtype=dtFloat) * fwhmRMSF
        statArr = np.ones((nY, nX), dtype="int") * (-1)

    # Remove redundant dimensions
    RMSFcube = np.squeeze(RMSFcube)
    fwhmRMSFArr = np.squeeze(fwhmRMSFArr)
    statArr = np.squeeze(statArr)

    return RMSFcube, phi2Arr, fwhmRMSFArr, statArr


def make_fake_data() -> FakeData:
    # Set up
    freq = np.arange(744, 1032, 1) * 1e6
    lsq = (2.998e8 / freq) ** 2
    rm = -100

    # Make fake data
    quArr = 1 * np.exp(2j * (np.radians(0) + rm * lsq))
    stokes_q = quArr.real
    stokes_u = quArr.imag
    weights = np.ones_like(freq)
    lsq_0 = np.average(lsq, weights=weights)
    phis = np.linspace(-500, 500, 10_000)

    return FakeData(
        freq=freq,
        lsq=lsq,
        rm=rm,
        stokes_q=stokes_q,
        stokes_u=stokes_u,
        weights=weights,
        lsq_0=lsq_0,
        phis=phis,
    )


def test_rmsynth() -> None:
    """Test the NUFFT RM-synthesis routine against DFT."""
    fake_data = make_fake_data()
    for eps in [1e-4, 1e-5, 1e-6, 1e-8]:
        tick = time()
        FDFcube = rmsynth_nufft(
            complex_pol_arr=fake_data.stokes_q + 1j * fake_data.stokes_u,
            lambda_sq_arr_m2=fake_data.lsq,
            phi_arr_radm2=fake_data.phis,
            weight_arr=fake_data.weights,
            lam_sq_0_m2=fake_data.lsq_0,
            eps=eps,
        )
        tock = time()
        msg = f"Time taken for NUFFT: {(tock - tick) * 1000:0.2f} ms"
        logger.info(msg)

        tick = time()
        FDFcube_old, lam0Sq_m2_old = do_rmsynth_planes_old(  # type: ignore[no-untyped-call]
            dataQ=fake_data.stokes_q,
            dataU=fake_data.stokes_u,
            lambdaSqArr_m2=fake_data.lsq,
            phiArr_radm2=fake_data.phis,
            weightArr=fake_data.weights,
            lam0Sq_m2=fake_data.lsq_0,
        )
        tock = time()
        msg = f"Time taken for DFT: {(tock - tick) * 1000:0.2f} ms"
        logger.info(msg)

        # fiNUFFT can't go below 1e-8 in precision!
        if eps == 1e-8:
            assert np.allclose(FDFcube, FDFcube_old, rtol=eps * 10, atol=eps * 10)
        else:
            assert np.allclose(FDFcube, FDFcube_old, rtol=eps, atol=eps)


def test_rmsf():
    """Test the NUFFT RMSF routine against DFT."""
    fake_data = make_fake_data()
    for eps in [1e-4, 1e-5, 1e-6, 1e-8]:
        tick = time()
        RMSFcube, phi2Arr, fwhmRMSFArr, statArr = get_rmsf_nufft(
            lambda_sq_arr_m2=fake_data.lsq,
            phi_arr_radm2=fake_data.phis,
            weight_arr=fake_data.weights,
            lam_sq_0_m2=fake_data.lsq_0,
            eps=eps,
        )
        tock = time()
        msg = f"Time taken for NUFFT: {(tock - tick) * 1000:0.2f} ms"
        logger.info(msg)

        tick = time()
        RMSFcube_old, phi2Arr_old, fwhmRMSFArr_old, statArr_old = get_rmsf_planes_old(  # type: ignore[no-untyped-call]
            lambdaSqArr_m2=fake_data.lsq,
            phiArr_radm2=fake_data.phis,
            weightArr=fake_data.weights,
            lam0Sq_m2=fake_data.lsq_0,
        )
        tock = time()
        msg = f"Time taken for DFT: {(tock - tick) * 1000:0.2f} ms"
        logger.info(msg)

        # fiNUFFT can't go below 1e-8 in precision!
        for name, new, old in zip(
            ["phi2Arr", "RMSFcube", "fwhmRMSFArr"],
            [phi2Arr, RMSFcube, fwhmRMSFArr],
            [phi2Arr_old, RMSFcube_old, fwhmRMSFArr_old],
        ):
            msg = f"Testing {name}"
            logger.info(msg)
            if eps == 1e-8:
                assert np.allclose(new, old, rtol=eps * 10, atol=eps * 10)
            else:
                msg = f"{new=},{old=}"
                logger.info(msg)
                assert np.allclose(new, old, rtol=eps * 2, atol=eps * 2)



# # =============================================================================#
# #                          MODEL DEFINITION FILE                              #
# # =============================================================================#
# import bilby
# import numpy as np
# from bilby.core.prior import Constraint, PriorDict


# # -----------------------------------------------------------------------------#
# # Function defining the model.                                                #
# #                                                                             #
# #  pDict       = Dictionary of parameters, created by parsing inParms, below. #
# #  lamSqArr_m2 = Array of lambda-squared values                               #
# #  quArr       = Complex array containing the Re and Im spectra.              #
# # -----------------------------------------------------------------------------#
# def model(pDict, lamSqArr_m2):
#     """

#     Two separate Faraday components with differential Faraday rotation
#     Double "Burn slab"
#     Averaged within the same telescope beam (i.e., unresolved)

#     Ref (for individual source component):
#     Burn (1966) Eq 18; with N >> (H_r/2H_z^0)^2
#     Sokoloff et al. (1998) Eq 3
#     O'Sullivan et al. (2012) Eq 9
#     Ma et al. (2019a) Eq 12

#     """

#     # Calculate the complex fractional q and u spectra
#     pArr1 = pDict["fracPol1"] * np.ones_like(lamSqArr_m2)
#     pArr2 = pDict["fracPol2"] * np.ones_like(lamSqArr_m2)

#     # fmt: off
#     quArr1 = pArr1 * np.exp( 2j * (np.radians(pDict["psi01_deg"]) +
#                                    (0.5*pDict["deltaRM1_radm2"] +
#                                     pDict["RM1_radm2"]) * lamSqArr_m2))
#     quArr2 = pArr2 * np.exp( 2j * (np.radians(pDict["psi02_deg"]) +
#                                    (0.5*pDict["deltaRM2_radm2"] +
#                                     pDict["RM2_radm2"]) * lamSqArr_m2))
#     quArr = (quArr1 * np.sin(pDict["deltaRM1_radm2"] * lamSqArr_m2) /
#              (pDict["deltaRM1_radm2"] * lamSqArr_m2) +
#              quArr2 * np.sin(pDict["deltaRM2_radm2"] * lamSqArr_m2) /
#              (pDict["deltaRM2_radm2"] * lamSqArr_m2))
#     # fmt: on

#     return quArr


# # -----------------------------------------------------------------------------#
# # Priors for the above model.                                                 #
# # See https://lscsoft.docs.ligo.org/bilby/prior.html for details.             #
# #                                                                             #
# # -----------------------------------------------------------------------------#
# def converter(parameters):
#     """
#     Function to convert between sampled parameters and constraint parameter.

#     Parameters
#     ----------
#     parameters: dict
#         Dictionary containing sampled parameter values, 'RM1_radm2', 'RM1_radm2',
#         'fracPol1', 'fracPol2'

#     Returns
#     -------
#     dict: Dictionary with constraint parameter 'delta_RM1_RM2_radm2' and 'sum_p1_p2' added.
#     """
#     converted_parameters = parameters.copy()
#     converted_parameters["delta_RM1_RM2_radm2"] = (
#         parameters["RM1_radm2"] - parameters["RM2_radm2"]
#     )
#     converted_parameters["sum_p1_p2"] = parameters["fracPol1"] + parameters["fracPol2"]
#     return converted_parameters


# priors = PriorDict(conversion_function=converter)

# priors["fracPol1"] = bilby.prior.Uniform(
#     minimum=0.0,
#     maximum=1.0,
#     name="fracPol1",
#     latex_label=r"$p_1$",
# )
# priors["fracPol2"] = bilby.prior.Uniform(
#     minimum=0.0,
#     maximum=1.0,
#     name="fracPol2",
#     latex_label=r"$p_2$",
# )

# priors["psi01_deg"] = bilby.prior.Uniform(
#     minimum=0,
#     maximum=180.0,
#     name="psi01_deg",
#     latex_label=r"$\psi_{0,1}$ (deg)",
#     boundary="periodic",
# )
# priors["psi02_deg"] = bilby.prior.Uniform(
#     minimum=0,
#     maximum=180.0,
#     name="psi02_deg",
#     latex_label=r"$\psi_{0,2}$ (deg)",
#     boundary="periodic",
# )

# priors["RM1_radm2"] = bilby.prior.Uniform(
#     minimum=-1100.0,
#     maximum=1100.0,
#     name="RM1_radm2",
#     latex_label="$\phi_1$ (rad m$^{-2}$)",
# )
# priors["RM2_radm2"] = bilby.prior.Uniform(
#     minimum=-1100.0,
#     maximum=1100.0,
#     name="RM2_radm2",
#     latex_label=r"$\phi_2$ (rad m$^{-2}$)",
# )
# priors["deltaRM1_radm2"] = bilby.prior.Uniform(
#     minimum=0.0,
#     maximum=100.0,
#     name="deltaRM1_radm2",
#     latex_label=r"$\Delta{RM,1}$ (rad m$^{-2}$))",
# )
# priors["deltaRM2_radm2"] = bilby.prior.Uniform(
#     minimum=0.0,
#     maximum=100.0,
#     name="deltaRM2_radm2",
#     latex_label=r"$\Delta{RM,2}$ (rad m$^{-2}$)",
# )
# priors["delta_RM1_RM2_radm2"] = Constraint(
#     minimum=0,
#     maximum=2200.0,
#     name="delta_RM1_RM2_radm2",
#     latex_label=r"$\Delta\phi_{1,2}$ (rad m$^{-2}$)",
# )
# priors["sum_p1_p2"] = Constraint(
#     minimum=0.0,
#     maximum=1,
#     name="sum_p1_p2",
#     latex_label=r"$p_1+p_2$)",
# )



"""RM-synthesis utils"""

from __future__ import annotations

import time
import warnings
from typing import Literal, NamedTuple, TypeVar

import finufft
import numpy as np
import polars as pl
from astropy.constants import c as speed_of_light
from astropy.stats import mad_std
from numpy.typing import NDArray
from scipy import stats
from tqdm.auto import trange
from uncertainties import unumpy

from rm_lite.utils.arrays import arange, nd_to_two_d, two_d_to_nd
from rm_lite.utils.fitting import FitResult, fit_fdf, fit_rmsf, fit_stokes_i_model
from rm_lite.utils.logging import logger


class FWHM(NamedTuple):
    fwhm_rmsf_radm2: float
    """The FWHM of the RMSF main lobe"""
    d_lambda_sq_max_m2: float
    """The maximum difference in lambda^2 values"""
    lambda_sq_range_m2: float
    """The range of lambda^2 values"""


class RMsynthResults(NamedTuple):
    """Results of the RM-synthesis calculation"""

    fdf_dirty_cube: NDArray[np.float64]
    """The Faraday dispersion function cube"""
    lam_sq_0_m2: float
    """The reference lambda^2 value"""


class RMSFResults(NamedTuple):
    """Results of the RMSF calculation"""

    rmsf_cube: NDArray[np.float64]
    """The RMSF cube"""
    phi_double_arr_radm2: NDArray[np.float64]
    """The (double length) Faraday depth array"""
    fwhm_rmsf_arr: NDArray[np.float64]
    """The FWHM of the RMSF main lobe"""
    fit_status_arr: NDArray[np.float64]
    """The status of the RMSF fit"""


class StokesData(NamedTuple):
    """Stokes parameters and errors"""

    complex_pol_arr: NDArray[np.complex128]
    """ Stokes Q and U array """
    complex_pol_error: NDArray[np.complex128]
    """ Stokes Q and U error array """
    freq_arr_hz: NDArray[np.float64]
    """ Frequency array in Hz """
    stokes_i_arr: NDArray[np.float64] | None = None
    """ Stokes I array """
    stokes_i_error_arr: NDArray[np.float64] | None = None
    """ Stokes I error array """
    stokes_i_model_arr: NDArray[np.float64] | None = None
    """ Stokes I model array """
    stokes_i_model_error: NDArray[np.float64] | None = None
    """ Stokes I model error array """


class FractionalSpectra(NamedTuple):
    stokes_i_model_arr: NDArray[np.float64]
    stokes_i_model_error: NDArray[np.float64]
    complex_pol_frac_arr: NDArray[np.complex128]
    complex_pol_frac_error: NDArray[np.complex128]
    fit_result: FitResult | None
    no_nan_idx: NDArray[np.bool_]


class TheoreticalNoise(NamedTuple):
    """Theoretical noise of the FDF"""

    fdf_error_noise: float
    """Theoretical noise of the FDF"""
    fdf_q_noise: float
    """Theoretical noise of the real FDF"""
    fdf_u_noise: float
    """Theoretical noise of the imaginary FDF"""

    def with_options(self, **kwargs):
        """Create a new TheoreticalNoise instance with keywords updated

        Returns:
            TheoreticalNoise: New TheoreticalNoise instance with updated attributes
        """
        # TODO: Update the signature to have the actual attributes to
        # help keep mypy and other linters happy
        as_dict = self._asdict()
        as_dict.update(kwargs)

        return TheoreticalNoise(**as_dict)


class FDFOptions(NamedTuple):
    """Options for RM-synthesis"""

    phi_max_radm2: float | None = None
    """ Maximum Faraday depth """
    d_phi_radm2: float | None = None
    """ Faraday depth resolution """
    n_samples: float | None = 10.0
    """ Number of samples """
    weight_type: Literal["variance", "uniform"] = "variance"
    """ Weight type """
    do_fit_rmsf: bool = False
    """ Fit RMSF """
    do_fit_rmsf_real: bool = False
    """ Fit real part of the RMSF """


def calc_mom2_fdf(
    complex_fdf_arr: NDArray[np.complex128], phi_arr_radm2: NDArray[np.float64]
) -> float:
    """
    Calculate the 2nd moment of the polarised intensity FDF. Can be applied to
    a clean component spectrum or a standard FDF
    """

    phi_weights = np.sum(np.abs(complex_fdf_arr))
    phi_mean = np.sum(phi_arr_radm2 * np.abs(complex_fdf_arr)) / phi_weights
    return float(
        np.sqrt(
            np.sum(np.power((phi_arr_radm2 - phi_mean), 2.0) * np.abs(complex_fdf_arr))
            / phi_weights
        )
    )


def create_fractional_spectra(
    stokes_data: StokesData,
    ref_freq_hz: float,
    fit_order: int = 2,
    fit_function: Literal["log", "linear"] = "log",
    n_error_samples: int = 10_000,
) -> FractionalSpectra:
    no_nan_idx = (
        np.isfinite(stokes_data.complex_pol_arr)
        & np.isfinite(stokes_data.complex_pol_error)
        & np.isfinite(stokes_data.freq_arr_hz)
    )

    # If no Stokes I at all, just return the 'fractional' spectra
    if (
        stokes_data.stokes_i_arr is None or stokes_data.stokes_i_error_arr is None
    ) and stokes_data.stokes_i_model_arr is None:
        logger.warning(
            "Stokes I array/errors or model not provided. No fractional polarization will be calculated."
        )
        stokes_i_arr = np.ones_like(stokes_data.complex_pol_arr.real)
        stokes_i_error_arr = np.zeros_like(stokes_data.complex_pol_error.real)

        return FractionalSpectra(
            stokes_i_model_arr=stokes_i_arr,
            stokes_i_model_error=stokes_i_error_arr,
            complex_pol_frac_arr=stokes_data.complex_pol_arr,
            complex_pol_frac_error=stokes_data.complex_pol_error,
            fit_result=None,
            no_nan_idx=no_nan_idx,
        )

    # Uncertainties doesn't support complex numbers, so we need to split the
    # Stokes Q and U arrays into real and imaginary parts
    stokes_q_uarray = unumpy.uarray(
        stokes_data.complex_pol_arr.real, stokes_data.complex_pol_error.real
    )
    stokes_u_uarray = unumpy.uarray(
        stokes_data.complex_pol_arr.imag, stokes_data.complex_pol_error.imag
    )

    # If a model is provided, use that to calculate the fractional spectra
    if stokes_data.stokes_i_model_arr is not None:
        logger.info("Using provided Stokes I model to calculate fractional spectra.")
        if stokes_data.stokes_i_model_error is None:
            msg = "If `stokes_i_model_arr` is provided, `stokes_i_model_error` must also be provided."
            raise ValueError(msg)

        stokes_i_model_uarray = unumpy.uarray(
            stokes_data.stokes_i_model_arr, stokes_data.stokes_i_model_error
        )

        frac_q_uarray = stokes_q_uarray / stokes_i_model_uarray
        frac_u_uarray = stokes_u_uarray / stokes_i_model_uarray

        stokes_q_frac_arr = unumpy.nominal_values(frac_q_uarray)
        stokes_u_frac_arr = unumpy.nominal_values(frac_u_uarray)

        stokes_q_frac_error_arr = unumpy.std_devs(frac_q_uarray)
        stokes_u_frac_error_arr = unumpy.std_devs(frac_u_uarray)

        stokes_qu_frac_arr = stokes_q_frac_arr + 1j * stokes_u_frac_arr
        stokes_qu_frac_error_arr = (
            stokes_q_frac_error_arr + 1j * stokes_u_frac_error_arr
        )

        return FractionalSpectra(
            stokes_i_model_arr=stokes_data.stokes_i_model_arr,
            stokes_i_model_error=stokes_data.stokes_i_model_error,
            complex_pol_frac_arr=stokes_qu_frac_arr.astype(np.complex128),
            complex_pol_frac_error=stokes_qu_frac_error_arr.astype(np.complex128),
            fit_result=None,
            no_nan_idx=no_nan_idx,
        )

    logger.info("Fitting Stokes I model to calculate fractional spectra.")
    if stokes_data.stokes_i_arr is None or stokes_data.stokes_i_error_arr is None:
        msg = "If `stokes_i_model_arr` is not provided, `stokes_i_arr` and `stokes_i_error_arr` must also be provided."
        raise ValueError(msg)

    # Flag out NaNs
    no_nan_idx = (
        no_nan_idx
        & np.isfinite(stokes_data.stokes_i_arr)
        & np.isfinite(stokes_data.stokes_i_error_arr)
    )
    logger.debug(f"{ref_freq_hz=}")

    # Apply flagging here since fitting will fail if NaNs are present
    fit_result = fit_stokes_i_model(
        freq_arr_hz=stokes_data.freq_arr_hz[no_nan_idx],
        ref_freq_hz=ref_freq_hz,
        stokes_i_arr=stokes_data.stokes_i_arr[no_nan_idx],
        stokes_i_error_arr=stokes_data.stokes_i_error_arr[no_nan_idx],
        fit_order=fit_order,
        fit_type=fit_function,
    )

    error_distribution = stats.multivariate_normal(
        mean=fit_result.popt,
        cov=fit_result.pcov,  # type: ignore[assignment,unused-ignore]
        allow_singular=True,
    )
    error_samples = error_distribution.rvs(n_error_samples)

    model_samples = np.array(
        [
            fit_result.stokes_i_model_func(
                stokes_data.freq_arr_hz / ref_freq_hz, *sample
            )
            for sample in error_samples
        ]
    )
    stokes_i_model_low, stokes_i_model_arr, stokes_i_model_high = np.nanpercentile(
        model_samples, [16, 50, 84], axis=0
    )
    stokes_i_model_error = np.abs(stokes_i_model_high - stokes_i_model_low)
    stokes_i_model_uarray = unumpy.uarray(
        stokes_i_model_arr,
        stokes_i_model_error,
    )

    stokes_q_frac_uarray = stokes_q_uarray / stokes_i_model_uarray
    stokes_u_frac_uarray = stokes_u_uarray / stokes_i_model_uarray

    stokes_q_frac_arr = np.array(unumpy.nominal_values(stokes_q_frac_uarray))
    stokes_u_frac_arr = np.array(unumpy.nominal_values(stokes_u_frac_uarray))
    stokes_q_frac_error_arr = np.array(unumpy.std_devs(stokes_q_frac_uarray))
    stokes_u_frac_error_arr = np.array(unumpy.std_devs(stokes_u_frac_uarray))

    assert len(stokes_data.stokes_i_arr) == len(stokes_q_frac_arr)
    assert len(stokes_data.stokes_i_arr) == len(stokes_u_frac_arr)
    assert len(stokes_data.stokes_i_arr) == len(stokes_q_frac_error_arr)
    assert len(stokes_data.stokes_i_arr) == len(stokes_u_frac_error_arr)

    complex_pol_arr = stokes_q_frac_arr + 1j * stokes_u_frac_arr
    complex_pol_error = stokes_q_frac_error_arr + 1j * stokes_u_frac_error_arr

    return FractionalSpectra(
        stokes_i_model_arr=stokes_i_model_arr,
        stokes_i_model_error=stokes_i_model_error,
        complex_pol_frac_arr=complex_pol_arr,
        complex_pol_frac_error=complex_pol_error,
        fit_result=fit_result,
        no_nan_idx=no_nan_idx,
    )


T = TypeVar("T", float, NDArray[np.float64])


def freq_to_lambda2(
    freq_hz: T,
) -> T:
    """Convert frequency to lambda^2.

    Args:
        freq_hz (float): Frequency in Hz

    Returns:
        float: Wavelength^2 in m^2
    """
    speed_of_light_m_s = float(speed_of_light.value)
    return (speed_of_light_m_s / freq_hz) ** 2.0  # type: ignore[no-any-return]


def lambda2_to_freq(lambda_sq_m2: T) -> T:
    """Convert lambda^2 to frequency.

    Args:
        lambda_sq_m2 (NDArray[np.float64]): Wavelength^2 in m^2

    Returns:
        NDArray[np.float64]: Frequency in Hz
    """
    speed_of_light_m_s = float(speed_of_light.value)
    return speed_of_light_m_s / np.sqrt(lambda_sq_m2)  # type: ignore[no-any-return]


def compute_theoretical_noise(
    complex_pol_frac_error: NDArray[np.complex128],
    weight_arr: NDArray[np.float64],
) -> TheoreticalNoise:
    weight_arr = np.nan_to_num(weight_arr, nan=0.0, posinf=0.0, neginf=0.0)
    complex_pol_frac_error_flagged = np.nan_to_num(
        complex_pol_frac_error, nan=0.0, posinf=0.0, neginf=0.0
    )
    fdf_complex_noise = np.sqrt(
        np.nansum(weight_arr**2 * complex_pol_frac_error_flagged**2)
        / (np.sum(weight_arr)) ** 2
    )

    fdf_error_noise = (fdf_complex_noise.real + fdf_complex_noise.imag) / 2
    return TheoreticalNoise(
        fdf_error_noise=fdf_error_noise,
        fdf_q_noise=fdf_complex_noise.real,
        fdf_u_noise=fdf_complex_noise.imag,
    )


class RMSynthParams(NamedTuple):
    """Parameters for RM-synthesis calculation"""

    lambda_sq_arr_m2: NDArray[np.float64]
    """ Wavelength^2 values in m^2 """
    lam_sq_0_m2: float
    """ Reference wavelength^2 value """
    phi_arr_radm2: NDArray[np.float64]
    """ Faraday depth values in rad/m^2 """
    weight_arr: NDArray[np.float64]
    """ Weight array """


class SigmaAdd(NamedTuple):
    """Sigma_add complexity metrics"""

    sigma_add: float
    """Sigma_add median value"""
    sigma_add_plus: float
    """Sigma_add upper quartile"""
    sigma_add_minus: float
    """Sigma_add lower quartile"""
    sigma_add_arrays: SigmaAddArrays
    """Sigma_add arrays"""


class StokesSigmaAdd(NamedTuple):
    """Stokes Sigma_add complexity metrics"""

    sigma_add_q: SigmaAdd
    """Sigma_add for Stokes Q"""
    sigma_add_u: SigmaAdd
    """Sigma_add for Stokes U"""
    sigma_add_p: SigmaAdd
    """Sigma_add for polarised intensity"""


class RMSFParams(NamedTuple):
    """RM spread function parameters"""

    rmsf_fwhm_theory: float
    """ Theoretical FWHM of the RMSF """
    rmsf_fwhm_meas: float
    """ Measured FWHM of the RMSF """
    phi_max: float
    """ Maximum Faraday depth """
    phi_max_scale: float
    """ Maximum Faraday depth scale """


def compute_rmsf_params(
    freq_arr_hz: NDArray[np.float64],
    weight_arr: NDArray[np.float64],
) -> RMSFParams:
    lambda_sq_arr_m2 = freq_to_lambda2(freq_arr_hz)
    # lam_sq_0_m2 is the weighted mean of lambda^2 distribution (B&dB Eqn. 32)
    # Calculate a global lam_sq_0_m2 value, ignoring isolated flagged voxels
    scale_factor = 1.0 / np.nansum(weight_arr)
    lam_sq_0_m2 = scale_factor * np.nansum(weight_arr * lambda_sq_arr_m2)
    if not np.isfinite(lam_sq_0_m2):
        lam_sq_0_m2 = np.nanmean(lambda_sq_arr_m2)

    lambda_sq_m2_max = np.nanmax(lambda_sq_arr_m2)
    lambda_sq_m2_min = np.nanmin(lambda_sq_arr_m2)
    delta_lambda_sq_m2 = np.median(np.abs(np.diff(lambda_sq_arr_m2)))

    rmsf_fwhm_theory = 3.8 / (lambda_sq_m2_max - lambda_sq_m2_min)
    phi_max = np.sqrt(3.0) / delta_lambda_sq_m2
    phi_max_scale = np.pi / lambda_sq_m2_min
    dphi = float(0.1 * rmsf_fwhm_theory)

    phi_arr_radm2 = make_phi_arr(phi_max * 10 * 2, dphi)

    rmsf_results = get_rmsf_nufft(
        lambda_sq_arr_m2=lambda_sq_arr_m2,
        phi_arr_radm2=phi_arr_radm2,
        weight_arr=weight_arr,
        lam_sq_0_m2=float(lam_sq_0_m2),
    )

    rmsf_fwhm_meas = float(rmsf_results.fwhm_rmsf_arr)

    return RMSFParams(
        rmsf_fwhm_theory=float(rmsf_fwhm_theory),
        rmsf_fwhm_meas=rmsf_fwhm_meas,
        phi_max=phi_max,
        phi_max_scale=float(phi_max_scale),
    )


def compute_rmsynth_params(
    freq_arr_hz: NDArray[np.float64],
    complex_pol_arr: NDArray[np.complex128],
    complex_pol_error: NDArray[np.complex128],
    fdf_options: FDFOptions,
) -> RMSynthParams:
    """Calculate the parameters for RM-synthesis.

    Args:
        freq_arr_hz (NDArray[np.float64]): Frequency array in Hz
        pol_arr (NDArray[np.complex128]): Complex polarisation array
        real_qu_error (NDArray[np.float64  |  np.float32]): Error in Stokes Q and U (real)
        fdf_options (FDFOptions): Options for RM-synthesis

    Raises:
        ValueError: If d_phi_radm2 is not provided and n_samples is None.

    Returns:
        RMSynthParams: Wavelength^2 values, reference wavelength^2, Faraday depth values, weight array
    """

    real_qu_error = np.abs(complex_pol_error.real + complex_pol_error.imag) / 2.0

    lambda_sq_arr_m2 = freq_to_lambda2(freq_arr_hz)

    fwhm_rmsf_radm2, d_lambda_sq_max_m2, lambda_sq_range_m2 = get_fwhm_rmsf(
        lambda_sq_arr_m2
    )

    if fdf_options.d_phi_radm2 is None and fdf_options.n_samples is not None:
        d_phi_radm2 = fwhm_rmsf_radm2 / fdf_options.n_samples
    elif fdf_options.d_phi_radm2 is not None:
        d_phi_radm2 = fdf_options.d_phi_radm2
    else:
        msg = "Either d_phi_radm2 or n_samples must be provided."
        raise ValueError(msg)

    if fdf_options.phi_max_radm2 is None:
        phi_max_radm2 = np.sqrt(3.0) / d_lambda_sq_max_m2
        phi_max_radm2 = max(
            phi_max_radm2, fwhm_rmsf_radm2 * 10.0
        )  # Force the minimum phiMax to 10 FWHM
    else:
        phi_max_radm2 = fdf_options.phi_max_radm2

    phi_arr_radm2 = make_phi_arr(phi_max_radm2, d_phi_radm2)

    logger.debug(
        f"phi = {phi_arr_radm2[0]:0.2f} to {phi_arr_radm2[-1]:0.2f} by {d_phi_radm2:0.2f} ({len(phi_arr_radm2)} chans)."
    )

    # Calculate the weighting as 1/sigma^2 or all 1s (uniform)
    if fdf_options.weight_type == "variance":
        if (real_qu_error == 0).all():
            real_qu_error = np.ones(len(real_qu_error))
        weight_arr = 1.0 / real_qu_error**2
    else:
        weight_arr = np.ones_like(freq_arr_hz)

    logger.debug(f"Weighting type: {fdf_options.weight_type}")

    mask = ~np.isfinite(complex_pol_arr)
    weight_arr[mask] = 0.0

    # lam_sq_0_m2 is the weighted mean of lambda^2 distribution (B&dB Eqn. 32)
    # Calculate a global lam_sq_0_m2 value, ignoring isolated flagged voxels
    scale_factor = 1.0 / np.nansum(weight_arr)
    lam_sq_0_m2 = float(scale_factor * np.nansum(weight_arr * lambda_sq_arr_m2))
    if not np.isfinite(lam_sq_0_m2):
        lam_sq_0_m2 = float(np.nanmean(lambda_sq_arr_m2))

    logger.debug(f"lam_sq_0_m2 = {lam_sq_0_m2:0.2f} m^2")

    return RMSynthParams(
        lambda_sq_arr_m2=lambda_sq_arr_m2,
        lam_sq_0_m2=lam_sq_0_m2,
        phi_arr_radm2=phi_arr_radm2,
        weight_arr=weight_arr,
    )


def make_phi_arr(
    phi_max_radm2: float,
    d_phi_radm2: float,
) -> NDArray[np.float64]:
    """Construct a Faraday depth array.

    Args:
        phi_max_radm2 (float): Maximum Faraday depth in rad/m^2
        d_phi_radm2 (float): Spacing in Faraday depth in rad/m^2

    Returns:
        NDArray[np.float64]: Faraday depth array in rad/m^2
    """
    # Faraday depth sampling. Zero always centred on middle channel
    n_chan_rm = int(np.round(abs((phi_max_radm2 - 0.0) / d_phi_radm2)) * 2.0 + 1.0)
    max_phi_radm2 = (n_chan_rm - 1.0) * d_phi_radm2 / 2.0
    return arange(
        start=-max_phi_radm2, stop=max_phi_radm2, step=d_phi_radm2, include_stop=True
    )


def make_double_phi_arr(
    phi_arr_radm2: NDArray[np.float64],
) -> NDArray[np.float64]:
    d_phi = phi_arr_radm2[1] - phi_arr_radm2[0]
    phi_max_radm2 = np.max(np.abs(phi_arr_radm2))
    return make_phi_arr(
        phi_max_radm2=phi_max_radm2 * 2 + d_phi,
        d_phi_radm2=d_phi,
    )


def get_fwhm_rmsf(
    lambda_sq_arr_m2: NDArray[np.float64],
) -> FWHM:
    """Calculate the FWHM of the RMSF.

    Args:
        lambda_sq_arr_m2 (NDArray[np.float64]): Wavelength^2 values in m^2
        super_resolution (bool, optional): Use Cotton+Rudnick superresolution. Defaults to False.

    Returns:
        fwhm_rmsf_arr: FWHM of the RMSF main lobe, maximum difference in lambda^2 values, range of lambda^2 values
    """
    lambda_sq_range_m2 = float(
        np.nanmax(lambda_sq_arr_m2) - np.nanmin(lambda_sq_arr_m2)
    )
    d_lambda_sq_max_m2 = np.nanmax(np.abs(np.diff(lambda_sq_arr_m2)))

    # Set the Faraday depth range
    fwhm_rmsf_radm2 = float(
        3.8 / lambda_sq_range_m2
    )  # Dickey+2019 theoretical RMSF width
    return FWHM(
        fwhm_rmsf_radm2=fwhm_rmsf_radm2,
        d_lambda_sq_max_m2=d_lambda_sq_max_m2,
        lambda_sq_range_m2=lambda_sq_range_m2,
    )


def rmsynth_nufft(
    complex_pol_arr: NDArray[np.complex128],
    lambda_sq_arr_m2: NDArray[np.float64],
    phi_arr_radm2: NDArray[np.float64],
    weight_arr: NDArray[np.float64],
    lam_sq_0_m2: float,
    eps: float = 1e-6,
) -> NDArray[np.complex128]:
    """Run RM-synthesis on a cube of Stokes Q and U data using the NUFFT method.

    Args:
        complex_pol_arr (NDArray[np.complex128]): Complex polarisation values (Q + iU)
        lambda_sq_arr_m2 (NDArray[np.float64]): Wavelength^2 values in m^2
        phi_arr_radm2 (NDArray[np.float64]): Faraday depth values in rad/m^2
        weight_arr (NDArray[np.float64]): Weight array
        lam_sq_0_m2 (Optional[float], optional): Reference wavelength^2 in m^2. Defaults to None.
        eps (float, optional): NUFFT tolerance. Defaults to 1e-6.

    Raises:
        ValueError: If the weight and lambda^2 arrays are not the same shape.
        ValueError: If the Stokes Q and U data arrays are not the same shape.
        ValueError: If the data dimensions are > 3.
        ValueError: If the data depth does not match the lambda^2 vector.

    Returns:
        NDArray[np.float64]: Dirty Faraday dispersion function cube
    """
    tick = time.time()
    msg = f"Running RM-synthesis using the NUFFTs over {len(phi_arr_radm2)} Faraday depth channels."
    logger.info(msg)
    flagged_weight_arr = np.nan_to_num(weight_arr, nan=0.0, posinf=0.0, neginf=0.0)

    # Sanity check on array sizes
    if flagged_weight_arr.shape != lambda_sq_arr_m2.shape:
        msg = f"Weight and lambda^2 arrays must be the same shape. Got {weight_arr.shape} and {lambda_sq_arr_m2.shape}"
        raise ValueError(msg)

    n_dims = len(complex_pol_arr.shape)
    if not n_dims <= 3:
        msg = f"Data dimensions must be <= 3. Got {n_dims}"
        raise ValueError(msg)

    if complex_pol_arr.shape[0] != lambda_sq_arr_m2.shape[0]:
        msg = f"Data depth does not match lambda^2 vector ({complex_pol_arr.shape[0]} vs {lambda_sq_arr_m2.shape[0]})."
        raise ValueError(msg)

    # Reshape the data arrays to 2 dimensions
    if n_dims == 1:
        complex_pol_arr_2d = np.reshape(complex_pol_arr, (complex_pol_arr.shape[0], 1))
    elif n_dims == 3:
        old_data_shape = complex_pol_arr.shape
        complex_pol_arr_2d = np.reshape(
            complex_pol_arr,
            (
                complex_pol_arr.shape[0],
                complex_pol_arr.shape[1] * complex_pol_arr.shape[2],
            ),
        )
    else:
        complex_pol_arr_2d = complex_pol_arr

    # Create a complex polarised cube, B&dB Eqns. (8) and (14)
    # Array has dimensions [nFreq, nY * nX]
    pol_cube = complex_pol_arr_2d * flagged_weight_arr[:, np.newaxis]

    # Check for NaNs (flagged data) in the cube & set to zero
    mask_cube = ~np.isfinite(pol_cube)
    pol_cube = np.nan_to_num(pol_cube, nan=0.0, posinf=0.0, neginf=0.0)

    # If full planes are flagged then set corresponding weights to zero
    mask_planes = np.sum(~mask_cube, axis=1)
    mask_planes = np.where(mask_planes == 0, 0, 1)
    flagged_weight_arr *= mask_planes

    # The K value used to scale each FDF spectrum must take into account
    # flagged voxels data in the datacube and can be position dependent
    weight_cube = np.invert(mask_cube) * flagged_weight_arr[:, np.newaxis]
    with np.errstate(divide="ignore", invalid="ignore"):
        scale_arr = np.true_divide(1.0, np.sum(weight_cube, axis=0))
        scale_arr[scale_arr == np.inf] = 0
        scale_arr = np.nan_to_num(scale_arr)

    # Clean up one cube worth of memory
    del weight_cube

    # Do the RM-synthesis on each plane
    # finufft must have matching dtypes, so complex64 matches float32
    exponent = (lambda_sq_arr_m2 - lam_sq_0_m2).astype(
        f"float{pol_cube.itemsize * 8 / 2:.0f}"
    )
    fdf_dirty_cube = (
        finufft.nufft1d3(
            x=exponent,
            c=np.ascontiguousarray(pol_cube.T),
            s=(phi_arr_radm2 * 2).astype(exponent.dtype),
            eps=eps,
            isign=-1,
        )
        * scale_arr[..., None]
    ).T

    # Check for pixels that have Re(FDF)=Im(FDF)=0. across ALL Faraday depths
    # These pixels will be changed to NaN in the output
    zeromap = np.all(fdf_dirty_cube == 0.0, axis=0)
    fdf_dirty_cube[..., zeromap] = np.nan + 1.0j * np.nan

    # Restore if 3D shape
    if n_dims == 3:
        fdf_dirty_cube = np.reshape(
            fdf_dirty_cube,
            (fdf_dirty_cube.shape[0], old_data_shape[1], old_data_shape[2]),
        )

    # Remove redundant dimensions in the FDF array
    tock = time.time()
    logger.info(f"NUFFT complete in {tock - tick:.3g} seconds.")
    return np.squeeze(fdf_dirty_cube)


def inverse_rmsynth_nufft(
    complex_fdf_arr: NDArray[np.complex128],
    lambda_sq_arr_m2: NDArray[np.float64],
    phi_arr_radm2: NDArray[np.float64],
    lam_sq_0_m2: float,
    eps: float = 1e-6,
) -> NDArray[np.complex128]:
    """Inverse RM-synthesis - FDF to Stokes Q and U in wavelength^2 space.

    Args:
        complex_fdf_arr (NDArray[np.complex128]): Complex polarisation array in Faraday depth space
        lambda_sq_arr_m2 (NDArray[np.float64]): Wavelength^2 values in m^2
        phi_arr_radm2 (NDArray[np.float64]): Faraday depth values in rad/m^2
        lam_sq_0_m2 (float): Reference wavelength^2 value
        eps (float, optional): NUFFT tolerance. Defaults to 1e-6.

    Raises:
        ValueError: If the Stokes Q and U data arrays are not the same shape.
        ValueError: If the data dimensions are > 3.
        ValueError: If the data depth does not match the lambda^2 vector.

    Returns:
        NDArray[np.float64]: Complex polarisation array in wavelength^2 space
    """

    # n_dims = len(fdf_q_arr.shape)
    # if not n_dims <= 3:
    #     msg = f"Data dimensions must be <= 3. Got {n_dims}"
    #     raise ValueError(msg)

    # if fdf_q_arr.shape[0] != phi_arr_radm2.shape[0]:
    #     msg = f"Data depth does not match Faraday depth vector ({fdf_q_arr.shape[0]} vs {phi_arr_radm2.shape[0]})."
    #     raise ValueError(msg)

    checks: list[tuple[bool, str]] = [
        (
            complex_fdf_arr.ndim <= 3,
            "Data dimensions must be <= 3.",
        ),
        (
            complex_fdf_arr.shape[0] == phi_arr_radm2.shape[0],
            f"Data depth does not match Faraday depth vector ({complex_fdf_arr.shape[0]} vs {phi_arr_radm2.shape[0]}).",
        ),
    ]
    for check, msg in checks:
        if not check:
            raise ValueError(msg)

    fdf_pol_cube_2d = nd_to_two_d(complex_fdf_arr)

    float_size = fdf_pol_cube_2d.itemsize * 8 / 2  # type: ignore[attr-defined,unused-ignore]
    exponent = (lambda_sq_arr_m2 - lam_sq_0_m2).astype(f"float{float_size:.0f}")
    pol_cube_inv = (
        finufft.nufft1d3(
            x=(phi_arr_radm2 * 2).astype(exponent.dtype),
            c=fdf_pol_cube_2d.T.astype(complex),  # type: ignore[attr-defined,unused-ignore]
            s=exponent,
            eps=eps,
            isign=1,
        )
    ).T

    # Restore if 3D shape
    if complex_fdf_arr.ndim == 3:
        pol_cube_inv = two_d_to_nd(pol_cube_inv, original_shape=complex_fdf_arr.shape)

    # Remove redundant dimensions in the FDF array
    return np.squeeze(pol_cube_inv).astype(np.complex128)


def get_rmsf_nufft(
    lambda_sq_arr_m2: NDArray[np.float64],
    phi_arr_radm2: NDArray[np.float64],
    weight_arr: NDArray[np.float64],
    lam_sq_0_m2: float,
    mask_arr: NDArray[np.bool_] | None = None,
    do_fit_rmsf: bool = False,
    do_fit_rmsf_real=False,
    eps: float = 1e-6,
) -> RMSFResults:
    """Compute the RMSF for a given set of lambda^2 values.

    Args:
        lambda_sq_arr_m2 (NDArray[np.float64]): Wavelength^2 values in m^2
        phi_arr_radm2 (NDArray[np.float64]): Faraday depth values in rad/m^2
        weight_arr (NDArray[np.float64]): Weight array
        lam_sq_0_m2 (float): Reference wavelength^2 value
        super_resolution (bool, optional): Use superresolution. Defaults to False.
        mask_arr (Optional[NDArray[np.float64]], optional): Mask array. Defaults to None.
        do_fit_rmsf (bool, optional): Fit the RMSF with a Gaussian. Defaults to False.
        do_fit_rmsf_real (bool, optional): Fit the *real* part of the. Defaults to False.
        eps (float, optional): NUFFT tolerance. Defaults to 1e-6.

    Raises:
        ValueError: If the wavelength^2 and weight arrays are not the same shape.
        ValueError: If the mask dimensions are > 3.
        ValueError: If the mask depth does not match the lambda^2 vector.

    Returns:
        RMSFResults: rmsf_cube, phi_double_arr_radm2, fwhm_rmsf_arr, fit_status_arr
    """
    phi_double_arr_radm2 = make_double_phi_arr(phi_arr_radm2)
    weight_arr = weight_arr.copy()
    weight_arr = np.nan_to_num(weight_arr, nan=0.0, posinf=0.0, neginf=0.0)

    # Set the mask array (default to 1D, no masked channels)
    if mask_arr is None:
        mask_arr = np.zeros_like(lambda_sq_arr_m2, dtype=bool)
        n_dimension = 1
    else:
        mask_arr = mask_arr.astype(bool)
        n_dimension = len(mask_arr.shape)

    # Sanity checks on array sizes
    if weight_arr.shape != lambda_sq_arr_m2.shape:
        msg = "wavelength^2 and weight arrays must be the same shape."
        raise ValueError(msg)

    if not n_dimension <= 3:
        msg = "mask dimensions must be <= 3."
        raise ValueError(msg)

    if mask_arr.shape[0] != lambda_sq_arr_m2.shape[0]:
        msg = f"Mask depth does not match lambda^2 vector ({mask_arr.shape[0]} vs {lambda_sq_arr_m2.shape[-1]})."
        raise ValueError(msg)

    # Reshape the mask array to 2 dimensions
    if n_dimension == 1:
        mask_arr = np.reshape(mask_arr, (mask_arr.shape[0], 1))
    elif n_dimension == 3:
        old_data_shape = mask_arr.shape
        mask_arr = np.reshape(
            mask_arr, (mask_arr.shape[0], mask_arr.shape[1] * mask_arr.shape[2])
        )
    num_pixels = mask_arr.shape[-1]

    # If full planes are flagged then set corresponding weights to zero
    flag_xy_sum = np.sum(mask_arr, axis=1)
    mskPlanes = np.where(flag_xy_sum == num_pixels, 0, 1)
    weight_arr *= mskPlanes

    fwhm_rmsf_radm2, _, _ = get_fwhm_rmsf(lambda_sq_arr_m2)
    # Calculate the RMSF at each pixel
    # The K value used to scale each RMSF must take into account
    # isolated flagged voxels data in the datacube
    weight_cube = np.invert(mask_arr) * weight_arr[:, np.newaxis]
    with np.errstate(divide="ignore", invalid="ignore"):
        scale_factor_arr = 1.0 / np.sum(weight_cube, axis=0)
        scale_factor_arr = np.nan_to_num(
            scale_factor_arr, nan=0.0, posinf=0.0, neginf=0.0
        )

    # Calculate the RMSF for each plane
    exponent = lambda_sq_arr_m2 - lam_sq_0_m2
    rmsf_cube = (
        finufft.nufft1d3(
            x=exponent,
            c=np.ascontiguousarray(weight_cube.T).astype(complex),
            s=(phi_double_arr_radm2[::-1] * 2).astype(exponent.dtype),
            eps=eps,
        )
        * scale_factor_arr[..., None]
    ).T

    # Clean up one cube worth of memory
    del weight_cube

    # Default to the analytical RMSF
    fwhm_rmsf_arr = np.ones(num_pixels) * fwhm_rmsf_radm2
    fit_status_arr = np.zeros(num_pixels, dtype=bool)

    # Fit the RMSF main lobe
    if do_fit_rmsf:
        logger.info("Fitting main lobe in each RMSF spectrum.")
        for i in trange(
            num_pixels, desc="Fitting RMSF by pixel", disable=num_pixels == 1
        ):
            try:
                fitted_rmsf = fit_rmsf(
                    rmsf_to_fit_arr=(
                        rmsf_cube[:, i].real
                        if do_fit_rmsf_real
                        else np.abs(rmsf_cube[:, i])
                    ),
                    phi_double_arr_radm2=phi_double_arr_radm2,
                    fwhm_rmsf_radm2=fwhm_rmsf_radm2,
                )
                fit_status = True
            except Exception as e:
                if num_pixels == 1:
                    raise e
                logger.error(f"Failed to fit RMSF at pixel {i}.")
                logger.error(e)
                logger.warning("Setting RMSF FWHM to default value.")
                fitted_rmsf = fwhm_rmsf_radm2
                fit_status = False

            fwhm_rmsf_arr[i] = fitted_rmsf
            fit_status_arr[i] = fit_status

    # Remove redundant dimensions
    rmsf_cube = np.squeeze(rmsf_cube)
    fwhm_rmsf_arr = np.squeeze(fwhm_rmsf_arr)
    fit_status_arr = np.squeeze(fit_status_arr)

    # Restore if 3D shape
    if n_dimension == 3:
        rmsf_cube = np.reshape(
            rmsf_cube, (rmsf_cube.shape[0], old_data_shape[1], old_data_shape[2])
        )
        fwhm_rmsf_arr = np.reshape(
            fwhm_rmsf_arr, (old_data_shape[1], old_data_shape[2])
        )
        fit_status_arr = np.reshape(
            fit_status_arr, (old_data_shape[1], old_data_shape[2])
        )

    return RMSFResults(
        rmsf_cube=rmsf_cube,
        phi_double_arr_radm2=phi_double_arr_radm2,
        fwhm_rmsf_arr=fwhm_rmsf_arr,
        fit_status_arr=fit_status_arr,
    )


fdf_params_schema = pl.Schema(
    {
        "fdf_error_mad": pl.Float64,
        "peak_pi_fit": pl.Float64,
        "peak_pi_error": pl.Float64,
        "peak_pi_fit_debias": pl.Float64,
        "peak_pi_fit_snr": pl.Float64,
        "peak_pi_fit_index": pl.Int64,
        "peak_rm_fit": pl.Float64,
        "peak_rm_fit_error": pl.Float64,
        "peak_q_fit": pl.Float64,
        "peak_u_fit": pl.Float64,
        "peak_pa_fit_deg": pl.Float64,
        "peak_pa_fit_deg_error": pl.Float64,
        "peak_pa0_fit_deg": pl.Float64,
        "peak_pa0_fit_deg_error": pl.Float64,
        "fit_function": pl.String,
        "lam_sq_0_m2": pl.Float64,
        "ref_freq_hz": pl.Float64,
        "fwhm_rmsf_radm2": pl.Float64,
        "fdf_error_noise": pl.Float64,
        "fdf_q_noise": pl.Float64,
        "fdf_u_noise": pl.Float64,
        "min_freq_hz": pl.Float64,
        "max_freq_hz": pl.Float64,
        "n_channels": pl.Int64,
        "median_d_freq_hz": pl.Float64,
        "frac_pol": pl.Float64,
        "frac_pol_error": pl.Float64,
        "sigma_add": pl.Float64,
        "sigma_add_minus": pl.Float64,
        "sigma_add_plus": pl.Float64,
    }
)
fdf_params_schema_df = fdf_params_schema.to_frame(eager=True)


def get_fdf_parameters(
    fdf_arr: NDArray[np.complex128],
    phi_arr_radm2: NDArray[np.float64],
    fwhm_rmsf_radm2: float,
    freq_arr_hz: NDArray[np.float64],
    complex_pol_arr: NDArray[np.complex128],
    complex_pol_error: NDArray[np.complex128],
    lambda_sq_arr_m2: NDArray[np.float64],
    lam_sq_0_m2: float,
    stokes_i_reference_flux: float,
    theoretical_noise: TheoreticalNoise,
    fit_function: Literal["log", "linear"],
    bias_correction_snr: float = 5.0,
) -> pl.DataFrame:
    """
    Measure standard parameters from a complex Faraday Dispersion Function.
    Currently this function assumes that the noise levels in the Stokes Q
    and U spectra are the same.
    Returns a dictionary containing measured parameters.
    """

    abs_fdf_arr = np.abs(fdf_arr)
    peak_pi_index = np.nanargmax(abs_fdf_arr)

    # Measure the RMS noise in the spectrum after masking the peak
    d_phi = phi_arr_radm2[1] - phi_arr_radm2[0]
    mask = np.ones_like(phi_arr_radm2, dtype=bool)
    mask[peak_pi_index] = False
    fwhm_rmsf_arr_pix = fwhm_rmsf_radm2 / d_phi
    for i in np.where(mask)[0]:
        start = int(i - fwhm_rmsf_arr_pix / 2)
        end = int(i + fwhm_rmsf_arr_pix / 2)
        mask[start : end + 2] = False

    # ignore mean of empty slice warning
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        fdf_error_mad = float(
            mad_std(np.concatenate([fdf_arr[mask].real, fdf_arr[mask].imag]))
        )

    n_good_phi = np.isfinite(fdf_arr).sum()
    lambda_sq_arr_m2_variance = (
        np.sum(lambda_sq_arr_m2**2.0) - np.sum(lambda_sq_arr_m2) ** 2.0 / n_good_phi
    ) / (n_good_phi - 1)

    good_chan_idx = np.isfinite(freq_arr_hz)
    n_good_chan = good_chan_idx.sum()

    if not (peak_pi_index > 0 and peak_pi_index < len(abs_fdf_arr) - 1):
        msg = "Peak index is not within the FDF array."
        raise ValueError(msg)

    peak_pi_fit, peak_rm_fit, _ = fit_fdf(
        fdf_to_fit_arr=abs_fdf_arr,
        phi_arr_radm2=phi_arr_radm2,
        fwhm_fdf_radm2=fwhm_rmsf_radm2,
    )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        peak_pi_fit_snr = peak_pi_fit / theoretical_noise.fdf_error_noise
    # In rare cases, a parabola can be fitted to the edge of the spectrum,
    # producing a unreasonably large RM and polarized intensity.
    # In these cases, everything should get NaN'd out.
    if np.abs(peak_rm_fit) > np.max(np.abs(phi_arr_radm2)):
        peak_rm_fit = np.nan
        peak_pi_fit = np.nan

    # Error on fitted Faraday depth (RM) is same as channel, but using fitted PI
    peak_rm_fit_err = (
        fwhm_rmsf_radm2 * theoretical_noise.fdf_error_noise / (2.0 * peak_pi_fit)
    )

    # Correct the peak for polarisation bias (POSSUM report 11)
    peak_pi_fit_debias = peak_pi_fit
    if peak_pi_fit_snr >= bias_correction_snr:
        peak_pi_fit_debias = np.sqrt(
            peak_pi_fit**2.0 - 2.3 * theoretical_noise.fdf_error_noise**2.0
        )

    # Calculate the polarisation angle from the fitted peak
    # Uncertainty from Eqn A.12 in Brentjens & De Bruyn 2005
    peak_pi_fit_index = np.interp(
        peak_rm_fit, phi_arr_radm2, np.arange(phi_arr_radm2.shape[-1], dtype="f4")
    )
    peak_u_fit = np.interp(peak_rm_fit, phi_arr_radm2, fdf_arr.imag)
    peak_q_fit = np.interp(peak_rm_fit, phi_arr_radm2, fdf_arr.real)
    peak_pa_fit_deg = 0.5 * np.degrees(np.arctan2(peak_u_fit, peak_q_fit)) % 180
    peak_pa_fit_deg_err = np.degrees(
        theoretical_noise.fdf_error_noise / (2.0 * peak_pi_fit)
    )

    # Calculate the derotated polarisation angle and uncertainty
    # Uncertainty from Eqn A.20 in Brentjens & De Bruyn 2005
    peak_pa0_fit_deg = (
        float(np.degrees(np.radians(peak_pa_fit_deg) - peak_rm_fit * lam_sq_0_m2))
        % 180.0
    )
    peak_pa0_fit_rad_err = np.sqrt(
        theoretical_noise.fdf_error_noise**2.0
        * n_good_phi
        / (4.0 * (n_good_phi - 2.0) * peak_pi_fit**2.0)
        * ((n_good_phi - 1) / n_good_phi + lam_sq_0_m2**2.0 / lambda_sq_arr_m2_variance)
    )
    peak_pa0_fit_deg_err = float(np.degrees(peak_pa0_fit_rad_err))

    stokes_sigma_add = measure_qu_complexity(
        freq_arr_hz=freq_arr_hz,
        complex_pol_arr=complex_pol_arr,
        complex_pol_error=complex_pol_error,
        frac_pol=peak_pi_fit_debias / stokes_i_reference_flux,
        psi0_deg=peak_pa0_fit_deg,
        rm_radm2=peak_rm_fit,
    )

    return fdf_params_schema_df.vstack(
        pl.DataFrame(
            {
                "fdf_error_mad": fdf_error_mad,
                "peak_pi_fit": peak_pi_fit,
                "peak_pi_error": theoretical_noise.fdf_error_noise,
                "peak_pi_fit_debias": peak_pi_fit_debias,
                "peak_pi_fit_snr": peak_pi_fit_snr,
                "peak_pi_fit_index": int(peak_pi_fit_index),
                "peak_rm_fit": peak_rm_fit,
                "peak_rm_fit_error": peak_rm_fit_err,
                "peak_q_fit": peak_q_fit,
                "peak_u_fit": peak_u_fit,
                "peak_pa_fit_deg": peak_pa_fit_deg,
                "peak_pa_fit_deg_error": peak_pa_fit_deg_err,
                "peak_pa0_fit_deg": peak_pa0_fit_deg,
                "peak_pa0_fit_deg_error": peak_pa0_fit_deg_err,
                "fit_function": fit_function,
                "lam_sq_0_m2": lam_sq_0_m2,
                "ref_freq_hz": lambda2_to_freq(lam_sq_0_m2),
                "fwhm_rmsf_radm2": fwhm_rmsf_radm2,
                "fdf_error_noise": theoretical_noise.fdf_error_noise,
                "fdf_q_noise": theoretical_noise.fdf_q_noise,
                "fdf_u_noise": theoretical_noise.fdf_u_noise,
                "min_freq_hz": freq_arr_hz[good_chan_idx].min(),
                "max_freq_hz": freq_arr_hz[good_chan_idx].max(),
                "n_channels": int(n_good_chan),
                "median_d_freq_hz": np.nanmedian(np.diff(freq_arr_hz[good_chan_idx])),
                "frac_pol": peak_pi_fit_debias / stokes_i_reference_flux,
                "frac_pol_error": theoretical_noise.fdf_error_noise
                / stokes_i_reference_flux,
                "sigma_add": stokes_sigma_add.sigma_add_p.sigma_add,
                "sigma_add_minus": stokes_sigma_add.sigma_add_p.sigma_add_minus,
                "sigma_add_plus": stokes_sigma_add.sigma_add_p.sigma_add_plus,
            }
        )
    )


def cdf_percentile(
    values: NDArray[np.float64], cdf: NDArray[np.float64], q=50.0
) -> float:
    """Return the value at a given percentile of a cumulative distribution function

    Args:
        values (NDArray[np.float64]): Array of values
        cdf (NDArray[np.float64]): Cumulative distribution function
        q (float, optional): Percentile. Defaults to 50.0.

    Returns:
        float: Interpolated value at the given percentile
    """
    return float(np.interp(q / 100.0, cdf, values))


class SigmaAddArrays(NamedTuple):
    pdf: NDArray[np.float64]
    """PDF array of the additional noise term"""
    cdf: NDArray[np.float64]
    """CDF array of the additional noise term"""
    sigma_add_arr: NDArray[np.float64]
    """Array of additional noise values"""


def calculate_sigma_add_arr(
    y_arr: NDArray[np.float64],
    dy_arr: NDArray[np.float64],
    median: float | None = None,
    noise: float | None = None,
    n_samples: int = 1000,
) -> SigmaAddArrays:
    # Measure the median and MADFM of the input data if not provided.
    # Used to overplot a normal distribution when debugging.
    if median is None:
        median = float(np.nanmedian(y_arr))
    if noise is None:
        noise = mad_std(y_arr)

    # Sample the PDF of the additional noise term from a limit near zero to
    # a limit of the range of the data, including error bars
    y_range = np.nanmax(y_arr + dy_arr) - np.nanmin(y_arr - dy_arr)
    sigma_add_arr = np.linspace(y_range / n_samples, y_range, n_samples)

    # Model deviation from Gaussian as an additional noise term.
    # Loop through the range of i additional noise samples and calculate
    # chi-squared and sum(ln(sigma_total)), used later to calculate likelihood.
    n_data = len(y_arr)

    # Calculate sigma_sq_tot for all sigma_add values
    sigma_sq_tot = dy_arr**2.0 + sigma_add_arr[:, None] ** 2.0

    # Calculate ln_sigma_sum_arr for all sigma_add values
    ln_sigma_sum_arr = np.nansum(np.log(np.sqrt(sigma_sq_tot)), axis=1)

    # Calculate chi_sq_arr for all sigma_add values
    chi_sq_arr = np.nansum((y_arr - median) ** 2.0 / sigma_sq_tot, axis=1)
    ln_prob_arr = (
        -np.log(sigma_add_arr)
        - n_data * np.log(2.0 * np.pi) / 2.0
        - ln_sigma_sum_arr
        - chi_sq_arr / 2.0
    )
    ln_prob_arr -= np.nanmax(ln_prob_arr)
    prob_arr = np.exp(ln_prob_arr)
    # Normalize the area under the PDF to be 1
    prob_arr /= np.nansum(prob_arr * np.diff(sigma_add_arr)[0])
    # Calculate the CDF
    cdf = np.cumsum(prob_arr) / np.nansum(prob_arr)

    return SigmaAddArrays(
        pdf=prob_arr,
        cdf=cdf,
        sigma_add_arr=sigma_add_arr,
    )


def calculate_sigma_add(
    y_arr: NDArray[np.float64],
    dy_arr: NDArray[np.float64],
    median: float | None = None,
    noise: float | None = None,
    n_samples: int = 1000,
) -> SigmaAdd:
    """Calculate the most likely value of additional scatter, assuming the
    input data is drawn from a normal distribution. The total uncertainty on
    each data point Y_i is modelled as dYtot_i**2 = dY_i**2 + dYadd**2."""

    sigma_add_arrays = calculate_sigma_add_arr(
        y_arr=y_arr,
        dy_arr=dy_arr,
        median=median,
        noise=noise,
        n_samples=n_samples,
    )

    # Calculate the mean of the distribution and the +/- 1-sigma limits
    sigma_add = cdf_percentile(
        values=sigma_add_arrays.sigma_add_arr, cdf=sigma_add_arrays.cdf, q=50.0
    )
    sigma_add_minus = cdf_percentile(
        values=sigma_add_arrays.sigma_add_arr, cdf=sigma_add_arrays.cdf, q=15.72
    )
    sigma_add_plus = cdf_percentile(
        values=sigma_add_arrays.sigma_add_arr, cdf=sigma_add_arrays.cdf, q=84.27
    )

    return SigmaAdd(
        sigma_add=sigma_add,
        sigma_add_minus=sigma_add_minus,
        sigma_add_plus=sigma_add_plus,
        sigma_add_arrays=sigma_add_arrays,
    )


def faraday_simple_spectrum(
    freq_arr_hz: NDArray[np.float64],
    frac_pol: float,
    psi0_deg: float,
    rm_radm2: float,
) -> NDArray[np.complex128]:
    """Create a simple Faraday spectrum with a single component.

    Args:
        freq_arr_hz (NDArray[np.float64]): Frequency array in Hz
        frac_pol (float): Fractional polarization
        psi0_deg (float): Initial polarization angle in degrees
        rm_radm2 (float): RM in rad/m^2

    Returns:
        NDArray[np.float64]: Complex polarization spectrum
    """
    lambda_sq_arr_m2 = freq_to_lambda2(freq_arr_hz)

    return frac_pol * np.exp(2j * (np.deg2rad(psi0_deg) + rm_radm2 * lambda_sq_arr_m2))


def measure_qu_complexity(
    freq_arr_hz: NDArray[np.float64],
    complex_pol_arr: NDArray[np.complex128],
    complex_pol_error: NDArray[np.complex128],
    frac_pol: float,
    psi0_deg: float,
    rm_radm2: float,
) -> StokesSigmaAdd:
    # Create a RM-thin model to subtract
    simple_model = faraday_simple_spectrum(
        freq_arr_hz=freq_arr_hz,
        frac_pol=frac_pol,
        psi0_deg=psi0_deg,
        rm_radm2=rm_radm2,
    )

    # Subtract the RM-thin model to create a residual q & u
    residual_qu = complex_pol_arr - simple_model

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)

        sigma_add_q = calculate_sigma_add(
            y_arr=residual_qu.real / complex_pol_error.real,
            dy_arr=np.ones_like(residual_qu.real),
            median=0.0,
            noise=1.0,
        )
        sigma_add_u = calculate_sigma_add(
            y_arr=residual_qu.imag / complex_pol_error.imag,
            dy_arr=np.ones_like(residual_qu.imag),
            median=0.0,
            noise=1.0,
        )

    sigma_add_p_arr = np.hypot(
        sigma_add_q.sigma_add_arrays.sigma_add_arr,
        sigma_add_u.sigma_add_arrays.sigma_add_arr,
    )
    sigma_add_p_pdf = np.hypot(
        sigma_add_q.sigma_add_arrays.pdf,
        sigma_add_u.sigma_add_arrays.pdf,
    )
    sigma_add_p_cdf = np.cumsum(sigma_add_p_pdf) / np.nansum(sigma_add_p_pdf)
    sigma_add_p_val = cdf_percentile(
        values=sigma_add_p_arr, cdf=sigma_add_p_cdf, q=50.0
    )
    sigma_add_p_minus = cdf_percentile(
        values=sigma_add_p_arr, cdf=sigma_add_p_cdf, q=15.72
    )
    sigma_add_p_plus = cdf_percentile(
        values=sigma_add_p_arr, cdf=sigma_add_p_cdf, q=84.27
    )
    sigma_add_p = SigmaAdd(
        sigma_add=sigma_add_p_val,
        sigma_add_minus=sigma_add_p_minus,
        sigma_add_plus=sigma_add_p_plus,
        sigma_add_arrays=SigmaAddArrays(
            pdf=sigma_add_p_pdf,
            cdf=sigma_add_p_cdf,
            sigma_add_arr=sigma_add_p_arr,
        ),
    )

    return StokesSigmaAdd(
        sigma_add_q=sigma_add_q,
        sigma_add_u=sigma_add_u,
        sigma_add_p=sigma_add_p,
    )


def measure_fdf_complexity(
    phi_arr_radm2: NDArray[np.float64], complex_fdf_arr: NDArray[np.complex128]
) -> float:
    # Second moment of clean component spectrum
    return calc_mom2_fdf(complex_fdf_arr=complex_fdf_arr, phi_arr_radm2=phi_arr_radm2)



# #!/usr/bin/env python3

# #                                                                             #
# # NAME:     rescale_I_model_3D.py                                             #
# #                                                                             #
# # PURPOSE:  Convert a 3D Stokes I model to a new reference frequency          #
# #                                                                             #
# # CREATED: 19-Mar-2024 by Cameron Van Eck
# #                                                                             #
# # =============================================================================#
# #                                                                             #
# # The MIT License (MIT)                                                       #
# #                                                                             #
# # Copyright (c) 2024 Cameron Van Eck                                          #
# #                                                                             #
# # Permission is hereby granted, free of charge, to any person obtaining a     #
# # copy of this software and associated documentation files (the "Software"),  #
# # to deal in the Software without restriction, including without limitation   #
# # the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
# # and/or sell copies of the Software, and to permit persons to whom the       #
# # Software is furnished to do so, subject to the following conditions:        #
# #                                                                             #
# # The above copyright notice and this permission notice shall be included in  #
# # all copies or substantial portions of the Software.                         #
# #                                                                             #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE #
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      #
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
# # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
# # DEALINGS IN THE SOFTWARE.                                                   #
# #                                                                             #
# # =============================================================================#


# # Steps: 1) Read in covariance matrix, coefficient maps, I ref.freq., lambda^2_0
# #       2a) Create fitDict, pixel-wise
# #       2b) Invoke rescaling function pixel-wise
# #       3) Drizzle output fitDict value into arrays
# #       4) Write new coefficient maps

# # Optional variations: user-specified reference frequency, unspecified
# #                      (make uniform across all pixels)

# import argparse
# import multiprocessing as mp
# import os
# from functools import partial

# import astropy.io.fits as pf
# import numpy as np

# from rm_lite.utils.fitting import FitResult, renormalize_StokesI_model


# def command_line():
#     """Handle invocation from the command line, parsing inputs and running
#     everything."""

#     # Help string to be shown using the -h option
#     descStr = """
#     Convert a Stokes I model to a new reference frequency. This changes the
#     model coefficients and their errors, but does not change (and does not
#     recalculate) the actual model spectrum.
#     The conversion of the coefficient uncertainties uses a first-order Taylor
#     approximation, so the uncertainties are only valid for relatively small
#     variations in reference frequency (depends a lot on the model, but 10%
#     seems to be a good rule of thumb).
#     The new reference frequency can either be from lambda^2_0 from an FDF cube,
#     which will create Stokes I maps that match the corresponding frequency,
#     a constant value given by the user, or unspecified (in which case the tool
#     will make all pixels have a common reference frequency at the mean input reference frequency).
#     The input files are assumed to be the products of do_fitIcube, with the usual filenames.
#     Outputs are new coefficient (and error) maps and reference frequency map.
#     """

#     # Parse the command line options
#     parser = argparse.ArgumentParser(
#         description=descStr, formatter_class=argparse.RawTextHelpFormatter
#     )

#     parser.add_argument(
#         "covar_file",
#         metavar="covariance.fits",
#         help="FITS cube Stokes I fit covariance matrices.",
#     )

#     parser.add_argument(
#         "-l",
#         dest="lambda2_0",
#         type=str,
#         default=None,
#         help="FDF cube from rmsynth3d. If given, will convert model to frequency matching the polarization products.",
#     )

#     parser.add_argument(
#         "-f",
#         dest="new_reffreq",
#         type=float,
#         default=None,
#         help="New reference frequency (in Hz). If given, forces all pixels to this frequency. Incompatible with -l",
#     )

#     parser.add_argument(
#         "-o",
#         dest="outname",
#         type=str,
#         default=None,
#         help="Output path+filename. If not given, defaults to input path+name (minus the covariance.fits).",
#     )
#     parser.add_argument(
#         "-n",
#         dest="num_cores",
#         type=int,
#         default=1,
#         help="Number of cores to use for multiprocessing. Default is 1.",
#     )

#     parser.add_argument(
#         "-w",
#         dest="overwrite",
#         action="store_true",
#         help="Overwrite existing files? [False].",
#     )

#     args = parser.parse_args()

#     if (args.new_reffreq is not None) and (args.lambda2_0 is not None):
#         raise Exception(
#             "Please do not set both -f and -l flags -- chose one or neither."
#         )

#     if not os.path.exists(args.covar_file):
#         raise Exception(
#             f"Cannot find covarance file at {args.covar_file}, please check filename/path."
#         )

#     if args.covar_file[-15:] != "covariance.fits":
#         raise Exception(
#             "Input covariance file name doesn't end in covariance.fits; this is required."
#         )

#     basename = args.covar_file[:-15]

#     if args.outname is None:
#         args.outname = basename

#     # Get data:
#     covar_map, old_reffreq_map, coeffs, header = read_data(basename)

#     # Create new frequency map:
#     if args.lambda2_0 is not None:
#         FDF_header = pf.getheader(args.lambda2_0)
#         lam0Sq_m2 = FDF_header["LAMSQ0"]
#         freq0 = 2.997924538e8 / np.sqrt(lam0Sq_m2)
#         new_freq_map = np.ones_like(old_reffreq_map) * freq0
#     elif args.new_reffreq is not None:
#         new_freq_map = np.ones_like(old_reffreq_map) * args.new_reffreq
#     else:
#         freq0 = np.nanmean(old_reffreq_map)
#         new_freq_map = np.ones_like(old_reffreq_map) * freq0

#     # Get fit function from header:
#     x = ["Fit model is" in card for card in header["HISTORY"]]
#     line = header["HISTORY"][np.where(x)[0][-1]]
#     fit_function = [x for x in line.split() if "polynomial" in x][0].split("-")[0]

#     new_freq_map, new_coeffs, new_errors = rescale_I_model_3D(
#         covar_map,
#         old_reffreq_map,
#         new_freq_map,
#         coeffs,
#         fit_function,
#         num_cores=args.num_cores,
#     )

#     write_new_parameters(
#         new_freq_map,
#         new_coeffs,
#         new_errors,
#         args.outname,
#         header,
#         overwrite=args.overwrite,
#     )


# def read_data(basename):
#     """Reads the covariance matrix map, (current) reference frequency map, and
#     (current) coefficient+error maps.
#     Input:
#         basename (str): file path and name up to 'covariance.fits'/'reffreq.fits', etc.
#     """

#     if not (
#         os.path.exists(basename + "coeff0.fits")
#         and os.path.exists(basename + "coeff0err.fits")
#     ):
#         raise Exception("Cannot find coeff0 map. At least coeff 0 map must exist.")

#     covar_file = basename + "covariance.fits"
#     covar_map = pf.getdata(covar_file)

#     freq_file = basename + "reffreq.fits"
#     old_reffreq_map, header = pf.getdata(freq_file, header=True)
#     # Grabs header from frequency map -- needed for fit function, and to use for
#     # writing out products. Better to have 2D map header to avoid fussing with
#     # extra axes.

#     # Get coefficient maps (without knowing how many there are)
#     # Reverse index order to match RM-Tools internal ordering (highest to lowest polynomial order)
#     coeffs = np.zeros((6, *old_reffreq_map.shape))
#     for i in range(6):
#         try:  # Keep trying higher orders
#             data = pf.getdata(basename + f"coeff{i}.fits")
#             coeffs[5 - i] = data
#         except FileNotFoundError:
#             break  # Once it runs out of valid coefficient maps, move on

#     return covar_map, old_reffreq_map, coeffs, header


# def rescale_I_pixel(data, fit_function):
#     covar, coeff, old_freq, new_freq = data
#     oldDict = {}  # Initialize a fitDict, which contains the relevant fit information
#     oldDict["reference_frequency_Hz"] = old_freq
#     oldDict["p"] = coeff
#     oldDict["pcov"] = covar
#     oldDict["fit_function"] = fit_function

#     old_result = FitResult(
#         params=coeff,
#         fitStatus=np.nan,  # Placeholder
#         chiSq=np.nan,  # Placeholder
#         chiSqRed=np.nan,  # Placeholder
#         AIC=np.nan,  # Placeholder
#         polyOrd=len(coeff) - 1,
#         nIter=0,  # Placeholder
#         reference_frequency_Hz=old_freq,
#         dof=np.nan,  # Placeholder
#         pcov=covar,
#         perror=np.zeros_like(coeff),  # Placeholder
#         fit_function=fit_function,
#     )

#     new_fit_result = renormalize_StokesI_model(old_result, new_freq)
#     return new_fit_result.params, new_fit_result.perror


# def rescale_I_model_3D(
#     covar_map, old_reffreq_map, new_freq_map, coeffs, fit_function="log", num_cores=1
# ):
#     """Rescale the Stokes I model parameters to a new reference frequency, for
#     an entire image (i.e., 3D pipeline products).

#     Inputs:
#         covar_map (4D array): covariance matrix map, such as produced by do_fitIcube.
#         old_reffreq_map (2D array): map of current reference frequency, such as produced by do_fitIcube.
#         new_freq_map (2D array): map of new reference frequencies.
#         coeffs (3D array): model parameter map (going from highest to lowest order)
#         coeff_errors (3D array): model parameter uncertainties map (highest to lowest order)

#     Returns:
#         new_freq_map (unchanged from input)
#         new_coeffs (3D array): maps of new model parameters (highest to lowest order)
#         new_errors (3D array): maps of new parameter uncertainties (highest to lowest)
#     """

#     # Initialize output arrays:
#     new_coeffs = np.zeros_like(coeffs)
#     new_errors = np.zeros_like(coeffs)
#     rs = old_reffreq_map.shape[
#         1
#     ]  # Get the length of a row, for array indexing later on.

#     # Set up inputs for parallelization:
#     # Input order is: covariance matrix, coefficient vector, old frequency, new frequency
#     inputs = list(
#         zip(
#             np.reshape(
#                 np.moveaxis(covar_map, (0, 1), (2, 3)), (old_reffreq_map.size, 6, 6)
#             ),
#             np.reshape(np.moveaxis(coeffs, 0, 2), (old_reffreq_map.size, 6)),
#             old_reffreq_map.flat,
#             new_freq_map.flat,
#         )
#     )
#     with mp.Pool(num_cores) as pool_:
#         results = pool_.map(
#             partial(rescale_I_pixel, fit_function=fit_function), inputs, chunksize=100
#         )

#     for i, (p, perror) in enumerate(results):
#         new_coeffs[:, i // rs, i % rs] = p
#         new_errors[:, i // rs, i % rs] = perror

#     return new_freq_map, new_coeffs, new_errors


# def write_new_parameters(
#     new_freq_map, new_coeffs, new_errors, out_basename, header, overwrite=False
# ):
#     """Write out new parameter/uncertainty maps to FITS files.
#     Inputs:
#         new_freq_map (unchanged from input)
#         new_coeffs (3D array): maps of new model parameters (highest to lowest order)
#         new_errors (3D array): maps of new parameter uncertainties (highest to lowest)
#         out_basename (str): base path+name of the files to be written out
#                             (will be postpended with 'newcoeff0.fits', etc.)

#     Returns: (nothing)
#         Writes out coefficient maps (newcoeff0.fits, etc.) and
#         coefficient errors (newcoeff0err.fits, etc.)
#     """

#     out_header = header.copy()
#     out_header["HISTORY"] = "Stokes I model rescaled to new reference frequency."
#     out_header["REFFREQ"] = (new_freq_map[0, 0], "Hz")
#     if "BUNIT" in out_header:
#         del out_header["BUNIT"]

#     # Work out highest order of polynomial:
#     max_order = np.sum(np.any(new_coeffs != 0.0, axis=(1, 2))) - 1

#     for i in range(max_order + 1):
#         pf.writeto(
#             out_basename + f"newcoeff{i}.fits",
#             new_coeffs[5 - i],
#             header=out_header,
#             overwrite=overwrite,
#         )
#         pf.writeto(
#             out_basename + f"newcoeff{i}err.fits",
#             new_errors[5 - i],
#             header=out_header,
#             overwrite=overwrite,
#         )


# # -----------------------------------------------------------------------------#
# if __name__ == "__main__":
#     command_line()
